import json
import os

from six import string_types
from six.moves.configparser import ConfigParser

import dtale.global_state as global_state
from dtale.utils import dict_merge

LOADED_CONFIG = None


def load_config_state(path):
    if not path:
        return None
    # load .ini file with properties specific to D-Tale
    config = ConfigParser()
    config.read(path)
    return config


def get_config():
    if LOADED_CONFIG:
        return LOADED_CONFIG
    ini_path = os.path.expandvars(
        os.environ.get("DTALE_CONFIG") or "$HOME/.config/dtale.ini"
    )
    if os.path.isfile(ini_path):
        return load_config_state(ini_path)
    return None


def set_config(path):
    global LOADED_CONFIG

    LOADED_CONFIG = load_config_state(path)


def get_config_val(config, defaults, prop, getter="get", section="show"):
    if config.has_option(section, prop):
        return getattr(config, getter)(section, prop)
    return defaults.get(prop)


def load_app_settings(config):
    if config is None:
        return
    curr_app_settings = global_state.get_app_settings()
    theme = get_config_val(config, curr_app_settings, "theme", section="app")
    pin_menu = get_config_val(
        config, curr_app_settings, "pin_menu", section="app", getter="getboolean"
    )
    language = get_config_val(config, curr_app_settings, "language", section="app")
    github_fork = get_config_val(
        config, curr_app_settings, "github_fork", section="app", getter="getboolean"
    )
    hide_shutdown = get_config_val(
        config, curr_app_settings, "hide_shutdown", section="app", getter="getboolean"
    )
    max_column_width = get_config_val(
        config, curr_app_settings, "max_column_width", section="app", getter="getint"
    )
    max_row_height = get_config_val(
        config, curr_app_settings, "max_row_height", section="app", getter="getint"
    )
    main_title = get_config_val(config, curr_app_settings, "main_title", section="app")
    main_title_font = get_config_val(
        config, curr_app_settings, "main_title_font", section="app"
    )
    query_engine = get_config_val(
        config, curr_app_settings, "query_engine", section="app"
    )
    hide_header_editor = get_config_val(
        config,
        curr_app_settings,
        "hide_header_editor",
        section="app",
        getter="getboolean",
    )
    hide_header_menu = get_config_val(
        config,
        curr_app_settings,
        "hide_header_menu",
        section="app",
        getter="getboolean",
    )
    hide_main_menu = get_config_val(
        config, curr_app_settings, "hide_main_menu", section="app", getter="getboolean"
    )
    hide_column_menus = get_config_val(
        config,
        curr_app_settings,
        "hide_column_menus",
        section="app",
        getter="getboolean",
    )
    hide_row_expanders = get_config_val(
        config,
        curr_app_settings,
        "hide_row_expanders",
        section="app",
        getter="getboolean",
    )
    lock_header_menu = get_config_val(
        config,
        curr_app_settings,
        "lock_header_menu",
        section="app",
        getter="getboolean",
    )
    enable_custom_filters = get_config_val(
        config,
        curr_app_settings,
        "enable_custom_filters",
        section="app",
        getter="getboolean",
    )
    enable_web_uploads = get_config_val(
        config,
        curr_app_settings,
        "enable_web_uploads",
        section="app",
        getter="getboolean",
    )
    open_custom_filter_on_startup = get_config_val(
        config,
        curr_app_settings,
        "open_custom_filter_on_startup",
        section="app",
        getter="getboolean",
    )
    open_predefined_filters_on_startup = get_config_val(
        config,
        curr_app_settings,
        "open_predefined_filters_on_startup",
        section="app",
        getter="getboolean",
    )
    hide_drop_rows = get_config_val(
        config, curr_app_settings, "hide_drop_rows", section="app", getter="getboolean"
    )

    global_state.set_app_settings(
        dict(
            theme=theme,
            pin_menu=pin_menu,
            language=language,
            github_fork=github_fork,
            hide_shutdown=hide_shutdown,
            max_column_width=max_column_width,
            max_row_height=max_row_height,
            main_title=main_title,
            main_title_font=main_title_font,
            query_engine=query_engine,
            open_custom_filter_on_startup=open_custom_filter_on_startup,
            open_predefined_filters_on_startup=open_predefined_filters_on_startup,
            hide_drop_rows=hide_drop_rows,
            hide_header_editor=hide_header_editor,
            lock_header_menu=lock_header_menu,
            hide_header_menu=hide_header_menu,
            hide_main_menu=hide_main_menu,
            hide_column_menus=hide_column_menus,
            hide_row_expanders=hide_row_expanders,
            enable_custom_filters=enable_custom_filters,
            enable_web_uploads=enable_web_uploads,
        )
    )


def load_chart_settings(config):
    if config is None:
        return
    curr_chart_settings = global_state.get_chart_settings()

    scatter_points = get_config_val(
        config, curr_chart_settings, "scatter_points", section="charts", getter="getint"
    )
    three_dimensional_points = get_config_val(
        config, curr_chart_settings, "3d_points", section="charts", getter="getint"
    )
    global_state.set_chart_settings(
        {"scatter_points": scatter_points, "3d_points": three_dimensional_points}
    )


def load_auth_settings(config):
    if config is None:
        return
    curr_auth_settings = global_state.get_auth_settings()
    active = get_config_val(
        config, curr_auth_settings, "active", section="auth", getter="getboolean"
    )
    username = get_config_val(config, curr_auth_settings, "username", section="auth")
    password = get_config_val(config, curr_auth_settings, "password", section="auth")

    global_state.set_auth_settings(
        dict(active=active, username=username, password=password)
    )


def build_show_options(options=None):
    defaults = dict(
        host=None,
        port=None,
        debug=False,
        subprocess=True,
        reaper_on=True,
        open_browser=False,
        notebook=False,
        force=False,
        ignore_duplicate=True,
        app_root=None,
        allow_cell_edits=True,
        inplace=False,
        drop_index=False,
        precision=2,
        show_columns=None,
        hide_columns=None,
        column_formats=None,
        nan_display=None,
        sort=None,
        locked=None,
        background_mode=None,
        range_highlights=None,
        vertical_headers=False,
        theme=None,
        hide_shutdown=None,
        column_edit_options=None,
        auto_hide_empty_columns=False,
        highlight_filter=False,
        hide_header_editor=None,
        lock_header_menu=None,
        hide_header_menu=None,
        hide_main_menu=None,
        hide_column_menus=None,
        hide_row_expanders=None,
        enable_custom_filters=None,
        enable_web_uploads=None,
        main_title=None,
        main_title_font=None,
    )
    config_options = {}
    config = get_config()
    if config and config.has_section("show"):
        config_options["host"] = get_config_val(config, defaults, "host")
        config_options["port"] = get_config_val(config, defaults, "port")
        config_options["debug"] = get_config_val(
            config, defaults, "debug", "getboolean"
        )
        config_options["subprocess"] = get_config_val(
            config, defaults, "subprocess", "getboolean"
        )
        config_options["reaper_on"] = get_config_val(
            config, defaults, "reaper_on", "getboolean"
        )
        config_options["open_browser"] = get_config_val(
            config, defaults, "open_browser", "getboolean"
        )
        config_options["notebook"] = get_config_val(
            config, defaults, "notebook", "getboolean"
        )
        config_options["force"] = get_config_val(
            config, defaults, "force", "getboolean"
        )
        config_options["ignore_duplicate"] = get_config_val(
            config, defaults, "ignore_duplicate", "getboolean"
        )
        config_options["app_root"] = get_config_val(config, defaults, "app_root")
        config_options["allow_cell_edits"] = get_config_val(
            config, defaults, "allow_cell_edits"
        )
        if isinstance(config_options["allow_cell_edits"], string_types):
            if config_options["allow_cell_edits"] == "True":
                config_options["allow_cell_edits"] = True
            elif config_options["allow_cell_edits"] == "False":
                config_options["allow_cell_edits"] = False
            else:
                config_options["allow_cell_edits"] = config_options[
                    "allow_cell_edits"
                ].split(",")

        config_options["inplace"] = get_config_val(
            config, defaults, "inplace", "getboolean"
        )
        config_options["drop_index"] = get_config_val(
            config, defaults, "drop_index", "getboolean"
        )
        config_options["precision"] = get_config_val(
            config, defaults, "precision", "getint"
        )
        config_options["show_columns"] = get_config_val(
            config, defaults, "show_columns"
        )
        if config_options["show_columns"]:
            config_options["show_columns"] = config_options["show_columns"].split(",")
        config_options["hide_columns"] = get_config_val(
            config, defaults, "hide_columns"
        )
        if config_options["hide_columns"]:
            config_options["hide_columns"] = config_options["hide_columns"].split(",")
        config_options["column_formats"] = get_config_val(
            config, defaults, "column_formats"
        )
        if config_options["column_formats"]:
            config_options["column_formats"] = json.loads(
                config_options["column_formats"]
            )
        config_options["nan_display"] = get_config_val(config, defaults, "nan_display")
        config_options["sort"] = get_config_val(config, defaults, "sort")
        if config_options["sort"]:
            config_options["sort"] = [
                tuple(sort.split("|")) for sort in config_options["sort"].split(",")
            ]
        config_options["locked"] = get_config_val(config, defaults, "locked")
        if config_options["locked"]:
            config_options["locked"] = config_options["locked"].split(",")
        config_options["background_mode"] = get_config_val(
            config, defaults, "background_mode"
        )
        config_options["range_highlights"] = get_config_val(
            config, defaults, "range_highlights"
        )
        if config_options["range_highlights"]:
            config_options["range_highlights"] = json.loads(
                config_options["range_highlights"]
            )
        config_options["vertical_headers"] = get_config_val(
            config, defaults, "vertical_headers", "getboolean"
        )
        config_options["column_edit_options"] = get_config_val(
            config, defaults, "column_edit_options"
        )
        if config_options["column_edit_options"]:
            config_options["column_edit_options"] = json.loads(
                config_options["column_edit_options"]
            )
        config_options["auto_hide_empty_columns"] = get_config_val(
            config, defaults, "auto_hide_empty_columns", "getboolean"
        )
        config_options["highlight_filter"] = get_config_val(
            config, defaults, "highlight_filter", "getboolean"
        )

    return dict_merge(defaults, config_options, options)


LOADED_CONFIG = get_config()
load_app_settings(LOADED_CONFIG)
load_auth_settings(LOADED_CONFIG)
load_chart_settings(LOADED_CONFIG)



import json as json
from logging import getLogger

import dash
from dash.dependencies import Input, Output, State
from six import PY3

from dtale.dash_application import dcc, html
import dtale.dash_application.custom_geojson as custom_geojson
import dtale.dash_application.drilldown_modal as drilldown_modal
from dtale.dash_application.exceptions import DtalePreventUpdate
import dtale.dash_application.extended_aggregations as extended_aggregations
import dtale.dash_application.saved_charts as saved_charts
import dtale.dash_application.lock_zoom as lock_zoom
import dtale.global_state as global_state
from dtale.charts.utils import (
    MAX_GROUPS,
    ZAXIS_CHARTS,
    NON_EXT_AGGREGATION,
    build_final_cols,
)
from dtale.code_export import build_final_chart_code
from dtale.dash_application.charts import build_chart, chart_url_params, valid_chart
from dtale.dash_application.layout.layout import (
    animate_styles,
    bar_input_style,
    base_layout,
    build_group_val_options,
    build_input_options,
    build_loc_mode_hover_children,
    build_map_options,
    build_candlestick_options,
    build_mapbox_style_options,
    build_pareto_options,
    build_proj_hover_children,
    build_slider_counts,
    build_label_value_options,
    charts_layout,
    colorscale_input_style,
    collapse_btn_text,
    get_yaxis_type_tabs,
    lock_zoom_style,
    main_inputs_and_group_val_display,
    show_chart_per_group,
    show_chart_per_y,
    get_group_types,
    show_input_handler,
    show_yaxis_ranges,
    bootstrap_checkbox_prop,
    get_num_cols,
)
from dtale.dash_application.layout.utils import show_style
from dtale.dash_application.utils import get_data_id
from dtale.translations import text
from dtale.query import handle_predefined, run_query
from dtale.utils import dict_merge, is_app_root_defined, make_list

logger = getLogger(__name__)


class DtaleDash(dash.Dash):
    """
    Wrapper class to dash.Dash to allow for abstraction of global state used for building the default layout.
    Additional state include stylesheets, JS files and styling for github demos.
    """

    def __init__(self, *args, **kwargs):
        server = kwargs.get("server")
        kwargs["external_stylesheets"] = [
            "/dtale/static/css/main.css",
            "/dtale/static/css/dash.css",
            "/dtale/static/css/github_fork.css",
        ]
        kwargs["external_scripts"] = [
            "/dtale/static/dash/components_bundle.js",
            "/dtale/static/dash/custom_bundle.js",
            "/dtale/static/dist/base_styles_bundle.js",
        ]

        app_root = server.config.get("APPLICATION_ROOT")
        if is_app_root_defined(app_root):

            def _prepend_app_root(v):
                return "{}{}".format(app_root, v)

            kwargs["requests_pathname_prefix"] = _prepend_app_root(
                kwargs["routes_pathname_prefix"]
            )
            kwargs["external_stylesheets"] = [
                _prepend_app_root(v) for v in kwargs["external_stylesheets"]
            ]
            kwargs["external_scripts"] = [
                _prepend_app_root(v) for v in kwargs["external_scripts"]
            ]
            kwargs["assets_url_path"] = _prepend_app_root("")
            kwargs["assets_external_path"] = _prepend_app_root("/assets")

        super(DtaleDash, self).__init__(*args, **kwargs)

    def interpolate_index(self, **kwargs):
        return base_layout(self.server.config.get("APPLICATION_ROOT"), **kwargs)


def add_dash(server):
    """
    Adds dash support to main D-Tale Flask process.

    :param server: main D-Tale Flask process
    :type server: :class:`flask:flask.Flask`
    :return: server with dash added
    :rtype: :class:`flask:flask.Flask`
    """

    dash_app = DtaleDash(
        server=server, routes_pathname_prefix="/dtale/charts/", eager_loading=True
    )

    # Since we're adding callbacks to elements that don't exist in the app.layout,
    # Dash will raise an exception to warn us that we might be
    # doing something wrong.
    # In this case, we're adding the elements through a callback, so we can ignore
    # the exception.
    dash_app.config.suppress_callback_exceptions = True
    dash_app.layout = html.Div(
        [dcc.Location(id="url", refresh=False), html.Div(id="popup-content")]
    )
    dash_app.scripts.config.serve_locally = True
    dash_app.css.config.serve_locally = True

    init_callbacks(dash_app)

    def _handle_error(error):
        """Replace the default handler with one that does not print anything"""
        import dtale.app as dtale_app

        if dtale_app.USE_COLAB:
            raise ValueError(
                "Currently executing in google colab which does not handle dash.PreventUpdate 204s"
            )
        return "", 204

    dash_app.server.errorhandler(DtalePreventUpdate)(_handle_error)

    return dash_app.server


def init_callbacks(dash_app):
    """
    Dynamically adds dash callbacks to dash-wrapped flask server

    :param dash_app: flask server with dash support enabled
    :type dash_app: :class:`flask:flask.Flask`
    :return: flask server with dash callbacks added
    :rtype: :class:`flask:flask.Flask`
    """

    @dash_app.callback(
        [
            Output("query-data", "data"),
            Output("query-input", "style"),
            Output("query-input", "title"),
            Output("load-input", "marks"),
        ],
        [Input("query-input", "value")],
        [
            State("query-data", "data"),
            State("load-input", "marks"),
            State("data-tabs", "value"),
        ],
    )
    def query_input(query, curr_query, curr_marks, data_id):
        """
        dash callback for storing valid pandas dataframe queries.  This acts as an intermediary between values typed
        by the user and values that are applied to pandas dataframes.  Most of the time what the user has typed is not
        complete and thus not a valid pandas dataframe query.

        :param query: query input
        :type query: str
        :param data_id: identifier for the data we are viewing
        :type data_id: string
        :param curr_query: current valid pandas dataframe query
        :return: tuple of (query (if valid), styling for query input (if invalid input), query input title (containing
        invalid query exception information)
        :rtype: tuple of (str, str, str)
        """
        try:
            data = handle_predefined(data_id)
            ctxt_vars = global_state.get_context_variables(data_id)
            df = run_query(data, query, ctxt_vars)
            return (
                query,
                {"line-height": "inherit"},
                "",
                build_slider_counts(df, data_id, query),
            )
        except BaseException as ex:
            return (
                curr_query,
                {"line-height": "inherit", "background-color": "pink"},
                str(ex),
                curr_marks,
            )

    @dash_app.callback(
        [
            Output("input-data", "data"),
            Output("x-dropdown", "options"),
            Output("y-single-dropdown", "options"),
            Output("y-multi-dropdown", "options"),
            Output("z-dropdown", "options"),
            Output("group-dropdown", "options"),
            Output("barsort-dropdown", "options"),
            Output("yaxis-dropdown", "options"),
            Output("standard-inputs", "style"),
            Output("map-inputs", "style"),
            Output("candlestick-inputs", "style"),
            Output("treemap-inputs", "style"),
            Output("funnel-inputs", "style"),
            Output("clustergram-inputs", "style"),
            Output("pareto-inputs", "style"),
            Output("histogram-inputs", "style"),
            Output("colorscale-input", "style"),
            Output("drilldown-input", "style"),
            Output("lock-zoom-btn", "style"),
            Output("open-extended-agg-modal", "style"),
            Output("selected-cleaners", "children"),
            Output("charts-filters-div", "style"),
            Output("stratified-group-dropdown", "style"),
        ],
        [
            Input("query-data", "modified_timestamp"),
            Input("extended-aggregations", "modified_timestamp"),
            Input("chart-tabs", "value"),
            Input("x-dropdown", "value"),
            Input("y-multi-dropdown", "value"),
            Input("y-single-dropdown", "value"),
            Input("z-dropdown", "value"),
            Input("group-dropdown", "value"),
            Input("group-type", "value"),
            Input("group-val-dropdown", "value"),
            Input("bins-val-input", "value"),
            Input("bin-type", "value"),
            Input("agg-dropdown", "value"),
            Input("window-input", "value"),
            Input("rolling-comp-dropdown", "value"),
            Input("load-input", "value"),
            Input("load-type-dropdown", "value"),
            Input("stratified-group-dropdown", "value"),
            Input("cleaners-dropdown", "value"),
            Input("dropna-checkbox", bootstrap_checkbox_prop()),
        ],
        [
            State("url", "pathname"),
            State("query-data", "data"),
            State("data-tabs", "value"),
            State("extended-aggregations", "data"),
        ],
    )
    def input_data(
        _ts,
        _ts2,
        chart_type,
        x,
        y_multi,
        y_single,
        z,
        group,
        group_type,
        group_val,
        bins_val,
        bin_type,
        agg,
        window,
        rolling_comp,
        load,
        load_type,
        stratified_group,
        cleaners,
        dropna,
        pathname,
        query,
        data_id,
        extended_aggregation,
    ):
        """
        dash callback for maintaining chart input state and column-based dropdown options.  This will guard against
        users selecting the same column for multiple axes.
        """
        y_val = make_list(y_single if chart_type in ZAXIS_CHARTS else y_multi)
        data_id = data_id or get_data_id(pathname)
        if group_val is not None:
            group_val = [json.loads(gv) for gv in group_val]

        inputs = dict(
            data_id=data_id,
            query=query,
            chart_type=chart_type,
            x=x,
            y=y_val,
            z=z,
            group=group,
            group_type=group_type or "groups",
            group_val=group_val if group else None,
            bins_val=bins_val if group else None,
            bin_type=bin_type or "width",
            agg=agg or "raw",
            window=window,
            rolling_comp=rolling_comp,
            load=load,
            load_type=load_type,
            stratified_group=stratified_group,
            cleaners=make_list(cleaners),
            dropna=True if dropna is None else dropna,
        )
        options = build_input_options(
            global_state.get_data(data_id),
            extended_aggregation=extended_aggregation,
            **inputs
        )
        (
            x_options,
            y_multi_options,
            y_single_options,
            z_options,
            group_options,
            barsort_options,
            yaxis_options,
            stratified_groups,
        ) = options
        show_map = chart_type == "maps"
        map_style = {} if show_map else {"display": "none"}
        show_cs = chart_type == "candlestick"
        cs_style = {} if show_cs else {"display": "none"}
        show_treemap = chart_type == "treemap"
        treemap_style = {} if show_treemap else {"display": "none"}
        show_funnel = chart_type == "funnel"
        funnel_style = {} if show_funnel else {"display": "none"}
        show_clustergram = chart_type == "clustergram"
        clustergram_style = {} if show_clustergram else {"display": "none"}
        show_pareto = chart_type == "pareto"
        pareto_style = {} if show_pareto else {"display": "none"}
        show_histogram = chart_type == "histogram"
        histogram_style = {} if show_histogram else {"display": "none"}
        standard_style = (
            {"display": "none"}
            if show_map
            or show_cs
            or show_treemap
            or show_funnel
            or show_clustergram
            or show_pareto
            or show_histogram
            else {}
        )
        cscale_style = colorscale_input_style(chart_type=chart_type)
        drilldown_toggle_style = show_style((agg or "raw") != "raw")
        return (
            inputs,
            x_options,
            y_single_options,
            y_multi_options,
            z_options,
            group_options,
            barsort_options,
            yaxis_options,
            standard_style,
            map_style,
            cs_style,
            treemap_style,
            funnel_style,
            clustergram_style,
            pareto_style,
            histogram_style,
            cscale_style,
            drilldown_toggle_style,
            lock_zoom_style(chart_type),
            show_style(chart_type not in NON_EXT_AGGREGATION and len(y_val)),
            (
                "({} Selected)".format(len(inputs["cleaners"]))
                if len(inputs["cleaners"])
                else ""
            ),
            {"display": "none"} if chart_type == "histogram" else {},
            {} if load_type == "stratified" else {"display": "none"},
        )

    @dash_app.callback(
        [
            Output("x-dropdown", "value"),
            Output("y-multi-dropdown", "value"),
            Output("y-single-dropdown", "value"),
            Output("z-dropdown", "value"),
            Output("group-dropdown", "value"),
            Output("query-input", "value"),
            Output("y-all-clicks", "data"),
        ],
        [Input("data-tabs", "value"), Input("y-select-all-btn", "n_clicks")],
        [
            State("input-data", "data"),
            State("y-all-clicks", "data"),
            State("x-dropdown", "value"),
            State("y-single-dropdown", "value"),
            State("z-dropdown", "value"),
            State("group-dropdown", "value"),
            State("query-input", "value"),
            State("y-multi-dropdown", "options"),
        ],
    )
    def update_data_selection(
        data_id,
        y_all_clicks,
        input_data,
        prev_y_all_clicks,
        x,
        y_single,
        z,
        group,
        query,
        y_multi_options,
    ):
        if y_all_clicks and y_all_clicks != prev_y_all_clicks:
            return (
                x,
                [o["value"] for o in y_multi_options],
                y_single,
                z,
                group,
                query,
                y_all_clicks,
            )
        if data_id == input_data["data_id"]:
            raise DtalePreventUpdate
        return None, None, None, None, None, None, None

    @dash_app.callback(
        [
            Output("map-input-data", "data"),
            Output("map-loc-dropdown", "options"),
            Output("map-lat-dropdown", "options"),
            Output("map-lon-dropdown", "options"),
            Output("map-val-dropdown", "options"),
            Output("map-loc-mode-input", "style"),
            Output("map-loc-input", "style"),
            Output("map-lat-input", "style"),
            Output("map-lon-input", "style"),
            Output("map-scope-input", "style"),
            Output("map-mapbox-style-input", "style"),
            Output("map-proj-input", "style"),
            Output("proj-hover", "style"),
            Output("proj-hover", "children"),
            Output("loc-mode-hover", "style"),
            Output("loc-mode-hover", "children"),
            Output("custom-geojson-input", "style"),
        ],
        [
            Input("map-type-tabs", "value"),
            Input("map-loc-mode-dropdown", "value"),
            Input("map-loc-dropdown", "value"),
            Input("map-lat-dropdown", "value"),
            Input("map-lon-dropdown", "value"),
            Input("map-val-dropdown", "value"),
            Input("map-scope-dropdown", "value"),
            Input("map-mapbox-style-dropdown", "value"),
            Input("map-proj-dropdown", "value"),
            Input("map-group-dropdown", "value"),
            Input("map-dropna-checkbox", bootstrap_checkbox_prop()),
            Input("geojson-dropdown", "value"),
            Input("featureidkey-dropdown", "value"),
        ],
        [State("data-tabs", "value")],
    )
    def map_data(
        map_type,
        loc_mode,
        loc,
        lat,
        lon,
        map_val,
        scope,
        style,
        proj,
        group,
        dropna,
        geojson,
        featureidkey,
        data_id,
    ):
        map_type = map_type or "choropleth"
        if map_type == "choropleth":
            map_data = dict(
                map_type=map_type,
                loc_mode=loc_mode,
                loc=loc,
                map_val=map_val,
                map_dropna=dropna,
            )
            if loc_mode == "geojson-id":
                map_data["geojson"] = geojson
                map_data["featureidkey"] = featureidkey
        elif map_type == "mapbox":
            map_data = dict(
                map_type=map_type,
                lat=lat,
                lon=lon,
                map_val=map_val,
                mapbox_style=style,
                map_dropna=dropna,
            )
        else:
            map_data = dict(
                map_type=map_type,
                lat=lat,
                lon=lon,
                map_val=map_val,
                scope=scope,
                proj=proj,
                map_dropna=dropna,
            )

        if group is not None:
            map_data["map_group"] = group
        df = global_state.get_data(data_id)
        loc_options, lat_options, lon_options, map_val_options = build_map_options(
            df, type=map_type, loc=loc, lat=lat, lon=lon, map_val=map_val
        )
        choro_style = {} if map_type == "choropleth" else {"display": "none"}
        coord_style = (
            {} if map_type in ["scattergeo", "mapbox"] else {"display": "none"}
        )
        scatt_style = {} if map_type == "scattergeo" else {"display": "none"}
        mapbox_style = {} if map_type == "mapbox" else {"display": "none"}
        proj_hover_style = (
            {"display": "none"} if proj is None else dict(borderBottom="none")
        )
        proj_hopver_children = build_proj_hover_children(proj)
        loc_mode_hover_style = (
            {"display": "none"} if loc_mode is None else dict(borderBottom="none")
        )
        loc_mode_children = build_loc_mode_hover_children(loc_mode)
        custom_geojson_link = (
            {}
            if map_type == "choropleth" and loc_mode == "geojson-id"
            else {"display": "none"}
        )
        return (
            map_data,
            loc_options,
            lat_options,
            lon_options,
            map_val_options,
            choro_style,
            choro_style,
            coord_style,
            coord_style,
            scatt_style,
            mapbox_style,
            scatt_style,
            proj_hover_style,
            proj_hopver_children,
            loc_mode_hover_style,
            loc_mode_children,
            custom_geojson_link,
        )

    @dash_app.callback(
        Output("map-mapbox-style-dropdown", "options"),
        [Input("mapbox-token-input", "value")],
    )
    def update_mapbox_token(token):
        from dtale.charts.utils import set_mapbox_token

        if token:
            set_mapbox_token(token)
        return build_mapbox_style_options()

    @dash_app.callback(
        [
            Output("candlestick-input-data", "data"),
            Output("candlestick-x-dropdown", "options"),
            Output("candlestick-open-dropdown", "options"),
            Output("candlestick-close-dropdown", "options"),
            Output("candlestick-high-dropdown", "options"),
            Output("candlestick-low-dropdown", "options"),
        ],
        [
            Input("candlestick-x-dropdown", "value"),
            Input("candlestick-open-dropdown", "value"),
            Input("candlestick-close-dropdown", "value"),
            Input("candlestick-high-dropdown", "value"),
            Input("candlestick-low-dropdown", "value"),
            Input("candlestick-group-dropdown", "value"),
            Input("candlestick-dropna-checkbox", bootstrap_checkbox_prop()),
        ],
        [State("data-tabs", "value")],
    )
    def cs_data_callback(
        cs_x, cs_open, cs_close, cs_high, cs_low, group, dropna, data_id
    ):
        cs_data = dict(
            cs_x=cs_x,
            cs_open=cs_open,
            cs_close=cs_close,
            cs_high=cs_high,
            cs_low=cs_low,
            cs_dropna=dropna,
        )
        if group is not None:
            cs_data["cs_group"] = group
        df = global_state.get_data(data_id)
        (
            x_options,
            close_options,
            open_options,
            low_options,
            high_options,
        ) = build_candlestick_options(
            df,
            cs_x=cs_x,
            cs_open=cs_open,
            cs_close=cs_close,
            cs_high=cs_high,
            cs_low=cs_low,
        )

        return (
            cs_data,
            x_options,
            open_options,
            close_options,
            high_options,
            low_options,
        )

    def label_value_callback(prop):
        def _callback(selected_value, selected_label, group, dropna, data_id, **kwargs):
            label_value_data = {
                "{}_value".format(prop): selected_value,
                "{}_label".format(prop): selected_label,
                "{}_dropna".format(prop): True if dropna is None else dropna,
            }
            if group is not None:
                label_value_data["{}_group".format(prop)] = group
            label_value_data = dict_merge(label_value_data, kwargs)
            df = global_state.get_data(data_id)
            value_options, label_options = build_label_value_options(
                df,
                selected_value=selected_value,
                selected_label=selected_label,
                all_value=prop == "clustergram",
            )
            return label_value_data, value_options, label_options

        return _callback

    def funnel_callback(
        selected_value, selected_label, group, dropna, stacked, data_id
    ):
        label_value_data, value_options, label_options = label_value_callback("funnel")(
            selected_value,
            selected_label,
            group,
            dropna,
            data_id,
            funnel_stacked=stacked,
        )
        return (
            label_value_data,
            value_options,
            label_options,
            show_style(len(make_list(group)) > 0),
        )

    def clustergram_callback(selected_value, selected_label, group, dropna, data_id):
        label_value_data, value_options, label_options = label_value_callback(
            "clustergram"
        )(selected_value, selected_label, group, dropna, data_id)
        return (label_value_data, value_options, label_options)

    @dash_app.callback(
        [
            Output("pareto-input-data", "data"),
            Output("pareto-x-dropdown", "options"),
            Output("pareto-bars-dropdown", "options"),
            Output("pareto-line-dropdown", "options"),
        ],
        [
            Input("pareto-x-dropdown", "value"),
            Input("pareto-bars-dropdown", "value"),
            Input("pareto-line-dropdown", "value"),
            Input("pareto-sort-dropdown", "value"),
            Input("pareto-dir-dropdown", "value"),
            Input("pareto-group-dropdown", "value"),
            Input("pareto-dropna-checkbox", bootstrap_checkbox_prop()),
        ],
        [State("data-tabs", "value")],
    )
    def pareto_data_callback(
        pareto_x,
        pareto_bars,
        pareto_line,
        pareto_sort,
        pareto_dir,
        group,
        dropna,
        data_id,
    ):
        pareto_data = dict(
            pareto_x=pareto_x,
            pareto_bars=pareto_bars,
            pareto_line=pareto_line,
            pareto_sort=pareto_sort,
            pareto_dir=pareto_dir,
            pareto_dropna=True if dropna is None else dropna,
        )
        if group is not None:
            pareto_data["pareto_group"] = group
        df = global_state.get_data(data_id)
        x_options, bars_options, line_options, _sort_options = build_pareto_options(
            df, x=pareto_x, bars=pareto_bars, line=pareto_line
        )

        return (pareto_data, x_options, bars_options, line_options)

    @dash_app.callback(
        [
            Output("histogram-input-data", "data"),
            Output("histogram-col-dropdown", "options"),
            Output("histogram-bins-div", "style"),
        ],
        [
            Input("histogram-col-dropdown", "value"),
            Input("histogram-type-tabs", "value"),
            Input("histogram-bins-input", "value"),
            Input("histogram-group-dropdown", "value"),
        ],
        [State("data-tabs", "value")],
    )
    def histogram_data_callback(
        histogram_col, histogram_type, histogram_bins, group, data_id
    ):
        histogram_data = dict(
            histogram_col=histogram_col,
            histogram_type=histogram_type,
            histogram_bins=histogram_bins,
        )
        if group is not None:
            histogram_data["histogram_group"] = group
        df = global_state.get_data(data_id)
        col_options = get_num_cols(df)

        return (
            histogram_data,
            col_options,
            {} if histogram_type == "bins" else {"display": "none"},
        )

    dash_app.callback(
        [
            Output("treemap-input-data", "data"),
            Output("treemap-value-dropdown", "options"),
            Output("treemap-label-dropdown", "options"),
        ],
        [
            Input("treemap-value-dropdown", "value"),
            Input("treemap-label-dropdown", "value"),
            Input("treemap-group-dropdown", "value"),
            Input("treemap-dropna-checkbox", bootstrap_checkbox_prop()),
        ],
        [State("data-tabs", "value")],
    )(label_value_callback("treemap"))

    dash_app.callback(
        [
            Output("funnel-input-data", "data"),
            Output("funnel-value-dropdown", "options"),
            Output("funnel-label-dropdown", "options"),
            Output("funnel-stack-input", "style"),
        ],
        [
            Input("funnel-value-dropdown", "value"),
            Input("funnel-label-dropdown", "value"),
            Input("funnel-group-dropdown", "value"),
            Input("funnel-dropna-checkbox", bootstrap_checkbox_prop()),
            Input("funnel-stack-toggle", "on"),
        ],
        [State("data-tabs", "value")],
    )(funnel_callback)

    dash_app.callback(
        [
            Output("clustergram-input-data", "data"),
            Output("clustergram-value-dropdown", "options"),
            Output("clustergram-label-dropdown", "options"),
        ],
        [
            Input("clustergram-value-dropdown", "value"),
            Input("clustergram-label-dropdown", "value"),
            Input("clustergram-group-dropdown", "value"),
            Input("clustergram-dropna-checkbox", bootstrap_checkbox_prop()),
        ],
        [State("data-tabs", "value")],
    )(clustergram_callback)

    @dash_app.callback(
        [
            Output("y-multi-input", "style"),
            Output("y-single-input", "style"),
            Output("z-input", "style"),
            Output("group-input", "style"),
            Output("rolling-inputs", "style"),
            Output("cpg-input", "style"),
            Output("cpy-input", "style"),
            Output("barmode-input", "style"),
            Output("barsort-input", "style"),
            Output("top-bars-input", "style"),
            Output("yaxis-input", "style"),
            Output("animate-input", "style"),
            Output("animate-by-input", "style"),
            Output("animate-by-dropdown", "options"),
            Output("trendline-input", "style"),
            Output("dropna-input", "style"),
        ],
        [Input("input-data", "modified_timestamp")],
        [State("input-data", "data"), State("url", "pathname")],
    )
    def input_toggles(_ts, inputs, pathname):
        """
        dash callback controlling showing/hiding of chart-specific inputs (for example z-axis) as well as chart
        formatting inputs (sorting for bars in bar chart, bar chart style (stacked) or y-axis ranges.
        """
        [chart_type, agg] = [inputs.get(p) for p in ["chart_type", "agg"]]
        show_input = show_input_handler(chart_type)

        y_multi_style = {"display": "block" if show_input("y", "multi") else "none"}
        y_single_style = {"display": "block" if show_input("y") else "none"}
        z_style = {"display": "block" if show_input("z") else "none"}
        group_style = {"display": "block" if show_input("group") else "none"}
        dropna_style = {"display": "block" if PY3 and show_input("group") else "none"}
        rolling_style = {"display": "inherit" if agg == "rolling" else "none"}
        cpg_style = {"display": "block" if show_chart_per_group(**inputs) else "none"}
        cpy_style = {"display": "block" if show_chart_per_y(**inputs) else "none"}
        bar_style, barsort_style = bar_input_style(**inputs)
        yaxis_style = {"display": "block" if show_yaxis_ranges(**inputs) else "none"}

        data_id = get_data_id(pathname)
        df = global_state.get_data(data_id)
        animate_style, animate_by_style, animate_opts = animate_styles(df, **inputs)
        trendline_style = dict(display="block" if chart_type == "scatter" else "none")
        return (
            y_multi_style,
            y_single_style,
            z_style,
            group_style,
            rolling_style,
            cpg_style,
            cpy_style,
            bar_style,
            barsort_style,
            barsort_style,
            yaxis_style,
            animate_style,
            animate_by_style,
            animate_opts,
            trendline_style,
            dropna_style,
        )

    @dash_app.callback(
        Output("chart-input-data", "data"),
        [
            Input("cpg-toggle", "on"),
            Input("cpy-toggle", "on"),
            Input("barmode-dropdown", "value"),
            Input("barsort-dropdown", "value"),
            Input("top-bars", "value"),
            Input("colorscale-picker", "colorscale"),
            Input("animate-toggle", "on"),
            Input("animate-by-dropdown", "value"),
            Input("trendline-dropdown", "value"),
            Input("yaxis-scale", "value"),
        ],
    )
    def chart_input_data(
        cpg,
        cpy,
        barmode,
        barsort,
        top_bars,
        colorscale,
        animate,
        animate_by,
        trendline,
        scale,
    ):
        """
        dash callback for maintaining selections in chart-formatting inputs
            - chart per group flag
            - bar chart mode
            - bar chart sorting
        """
        return dict(
            cpg=cpg,
            cpy=cpy,
            barmode=barmode,
            barsort=barsort,
            top_bars=top_bars,
            colorscale=colorscale,
            animate=animate,
            animate_by=animate_by,
            trendline=trendline,
            scale=scale,
        )

    @dash_app.callback(Output("load-btn", "style"), [Input("auto-load-toggle", "on")])
    def load_style(auto_load):
        return dict(display="block" if not auto_load else "none")

    @dash_app.callback(
        [Output("collapse-data", "is_open"), Output("collapse-data-btn", "children")],
        [Input("collapse-data-btn", "n_clicks")],
        [State("collapse-data", "is_open")],
    )
    def collapse_data_input(n, is_open):
        final_is_open = is_open
        if n:
            final_is_open = not is_open
        return final_is_open, collapse_btn_text(final_is_open, text("Data Selection"))

    @dash_app.callback(
        [
            Output("collapse-cleaners", "is_open"),
            Output("collapse-cleaners-btn", "children"),
        ],
        [Input("collapse-cleaners-btn", "n_clicks")],
        [State("collapse-cleaners", "is_open")],
    )
    def collapse_cleaners_input(n, is_open):
        final_is_open = is_open
        if n:
            final_is_open = not is_open
        return final_is_open, collapse_btn_text(final_is_open, text("Cleaners"))

    @dash_app.callback(
        [
            Output("chart-content", "children"),
            Output("last-chart-input-data", "data"),
            Output("range-data", "data"),
            Output("chart-code", "value"),
            Output("yaxis-type", "children"),
            Output("load-clicks", "data"),
            Output("save-btn", "style"),
            Output("agg-dropdown", "disabled"),
            Output("extended-aggregation-tooltip", "children"),
            Output("ext-agg-warning", "style"),
            Output("export-all-chart-btn", "href"),
            Output("export-all-chart-btn", "style"),
        ],
        # Since we use the data prop in an output,
        # we cannot get the initial data on load with the data prop.
        # To counter this, you can use the modified_timestamp
        # as Input and the data as State.
        # This limitation is due to the initial None callbacks
        # https://github.com/plotly/dash-renderer/pull/81
        [
            Input("input-data", "modified_timestamp"),
            Input("chart-input-data", "modified_timestamp"),
            Input("yaxis-data", "modified_timestamp"),
            Input("map-input-data", "modified_timestamp"),
            Input("candlestick-input-data", "modified_timestamp"),
            Input("treemap-input-data", "modified_timestamp"),
            Input("funnel-input-data", "modified_timestamp"),
            Input("clustergram-input-data", "modified_timestamp"),
            Input("pareto-input-data", "modified_timestamp"),
            Input("histogram-input-data", "modified_timestamp"),
            Input("extended-aggregations", "modified_timestamp"),
            Input("load-btn", "n_clicks"),
        ],
        [
            State("input-data", "data"),
            State("chart-input-data", "data"),
            State("yaxis-data", "data"),
            State("map-input-data", "data"),
            State("candlestick-input-data", "data"),
            State("treemap-input-data", "data"),
            State("funnel-input-data", "data"),
            State("clustergram-input-data", "data"),
            State("pareto-input-data", "data"),
            State("histogram-input-data", "data"),
            State("last-chart-input-data", "data"),
            State("auto-load-toggle", "on"),
            State("load-clicks", "data"),
            State("extended-aggregations", "data"),
        ],
    )
    def on_data(
        _ts1,
        _ts2,
        _ts3,
        _ts4,
        _ts5,
        _ts6,
        _ts7,
        _ts8,
        _ts9,
        _ts10,
        _ts11,
        load_clicks,
        inputs,
        chart_inputs,
        yaxis_data,
        map_data,
        cs_data,
        treemap_data,
        funnel_data,
        clustergram_data,
        pareto_data,
        histogram_data,
        last_chart_inputs,
        auto_load,
        prev_load_clicks,
        ext_aggs,
    ):
        """
        dash callback controlling the building of dash charts
        """
        duplicate_yaxis_update = False
        try:
            from dash import ctx

            if ctx.triggered_id == "yaxis-data":
                yaxis_type = (yaxis_data or {}).get("type")
                yaxis_selections = (yaxis_data or {}).get("data", {})
                populated_range = next(
                    (
                        y_range
                        for y_range in yaxis_selections.values()
                        if y_range["min"] is not None or y_range["max"] is not None
                    ),
                    None,
                )
                if yaxis_type == "default" and populated_range is None:
                    duplicate_yaxis_update = True
        except BaseException:
            logger.exception(
                "Could not determine if yaxis data is causing a duplicate render"
            )

        if duplicate_yaxis_update:
            raise DtalePreventUpdate

        all_inputs = dict_merge(
            inputs,
            chart_inputs,
            dict(yaxis=yaxis_data or {}),
            map_data,
            cs_data,
            treemap_data,
            funnel_data,
            clustergram_data,
            pareto_data,
            histogram_data,
            (
                dict(extended_aggregation=ext_aggs or [])
                if inputs.get("chart_type") not in NON_EXT_AGGREGATION
                else {}
            ),
        )
        if not auto_load and load_clicks == prev_load_clicks:
            raise DtalePreventUpdate
        if all_inputs == last_chart_inputs:
            raise DtalePreventUpdate
        if is_app_root_defined(dash_app.server.config.get("APPLICATION_ROOT")):
            all_inputs["app_root"] = dash_app.server.config["APPLICATION_ROOT"]
        charts, range_data, code, export_all_href = build_chart(**all_inputs)
        agg_disabled = len(ext_aggs) > 0
        ext_agg_tt = text("ext_agg_desc")
        ext_agg_warning = show_style(agg_disabled)
        if agg_disabled:
            ext_agg_tt = html.Div(
                [
                    html.Span(text("ext_agg_desc")),
                    html.Br(),
                    html.Ul(
                        [
                            html.Li(
                                extended_aggregations.build_extended_agg_desc(ext_agg),
                                className="mb-0",
                            )
                            for ext_agg in ext_aggs
                        ]
                    ),
                ]
            )
        final_cols = build_final_cols(
            make_list(inputs.get("y")),
            inputs.get("z"),
            inputs.get("agg"),
            ext_aggs if inputs.get("chart_type") not in NON_EXT_AGGREGATION else [],
        )
        return (
            charts,
            all_inputs,
            range_data,
            build_final_chart_code(code),
            get_yaxis_type_tabs(final_cols),
            load_clicks,
            dict(display="block" if valid_chart(**all_inputs) else "none"),
            agg_disabled,
            ext_agg_tt,
            ext_agg_warning,
            export_all_href,
            dict(display="none" if export_all_href == "" else "block"),
        )

    def get_default_range(range_data, y, max=False):
        if max:
            return next(
                iter(
                    sorted(
                        [range_data[y2] for y2 in y if y2 in range_data], reverse=True
                    )
                ),
                None,
            )
        return next(
            iter(sorted([range_data[y2] for y2 in y if y2 in range_data])), None
        )

    @dash_app.callback(
        [
            Output("yaxis-min-input", "value"),
            Output("yaxis-max-input", "value"),
            Output("yaxis-dropdown", "style"),
            Output("yaxis-min-label", "style"),
            Output("yaxis-min-input", "style"),
            Output("yaxis-max-label", "style"),
            Output("yaxis-max-input", "style"),
            Output("yaxis-type-div", "style"),
        ],
        [Input("yaxis-type", "value"), Input("yaxis-dropdown", "value")],
        [
            State("input-data", "data"),
            State("yaxis-data", "data"),
            State("range-data", "data"),
            State("extended-aggregations", "data"),
        ],
    )
    def yaxis_min_max_values(
        yaxis_type, yaxis, inputs, yaxis_inputs, range_data, ext_aggs
    ):
        """
        dash callback controlling values for selected y-axis in y-axis range editor
        """
        y = make_list(inputs.get("y"))
        final_cols = build_final_cols(
            y,
            inputs.get("z"),
            inputs.get("agg"),
            ext_aggs if inputs.get("chart_type") not in NON_EXT_AGGREGATION else [],
        )
        dd_style = dict(
            display="block" if yaxis_type == "multi" and len(final_cols) > 1 else "none"
        )
        type_style = (
            {"borderRadius": "0 0.25rem 0.25rem 0"} if yaxis_type == "default" else None
        )
        min_max_style = (
            "none"
            if (yaxis_type == "default") or (yaxis_type == "multi" and yaxis is None)
            else "block"
        )
        label_style = dict(display=min_max_style)
        input_style = {"lineHeight": "inherit", "display": min_max_style}
        curr_min, curr_max = (None, None)
        range_min, range_max = ((range_data or {}).get(p) or {} for p in ["min", "max"])
        if yaxis:
            curr_vals = (yaxis_inputs or {}).get("data", {}).get(yaxis) or {}
            curr_min = curr_vals.get("min") or range_min.get(yaxis)
            curr_max = curr_vals.get("max") or range_max.get(yaxis)
        elif yaxis_type == "single":
            curr_vals = (yaxis_inputs or {}).get("data", {}).get("all") or {}
            curr_min = curr_vals.get("min")
            if curr_min is None:
                curr_min = get_default_range(range_min, final_cols)
            curr_max = curr_vals.get("max")
            if curr_max is None:
                curr_max = get_default_range(range_max, final_cols, max=True)
        return (
            curr_min,
            curr_max,
            dd_style,
            label_style,
            input_style,
            label_style,
            input_style,
            type_style,
        )

    @dash_app.callback(
        Output("yaxis-data", "data"),
        [
            Input("yaxis-type", "value"),
            Input("yaxis-min-input", "value"),
            Input("yaxis-max-input", "value"),
        ],
        [
            State("yaxis-dropdown", "value"),
            State("yaxis-dropdown", "options"),
            State("yaxis-data", "data"),
            State("range-data", "data"),
            State("input-data", "data"),
            State("extended-aggregations", "data"),
        ],
    )
    def update_yaxis_data(
        yaxis_type,
        yaxis_min,
        yaxis_max,
        yaxis,
        yaxes,
        yaxis_data,
        range_data,
        inputs,
        ext_aggs,
    ):
        """
        dash callback controlling updates to y-axis range state
        """
        yaxis_data = yaxis_data or dict(data={})
        yaxis_data["type"] = yaxis_type
        yaxis = yaxis or yaxes[0]["value"] if len(yaxes) else None
        yaxis_name = "all" if yaxis_type == "single" else yaxis
        if yaxis_name == "all":
            final_cols = build_final_cols(
                make_list(inputs.get("y")),
                inputs.get("z"),
                inputs.get("agg"),
                ext_aggs if inputs.get("chart_type") not in NON_EXT_AGGREGATION else [],
            )
            mins = range_data.get("min", {})
            maxs = range_data.get("max", {})
            range_min = get_default_range(mins, final_cols)
            range_max = get_default_range(maxs, final_cols, max=True)
        elif yaxis is None or range_data is None:
            raise DtalePreventUpdate
        else:
            range_min, range_max = (
                range_data[p].get(yaxis_name) for p in ["min", "max"]
            )

        if yaxis_name in yaxis_data["data"]:
            if (yaxis_min, yaxis_max) == (range_min, range_max):
                del yaxis_data["data"][yaxis_name]
            else:
                yaxis_data["data"][yaxis_name] = dict(min=yaxis_min, max=yaxis_max)
        else:
            if (yaxis_min, yaxis_max) != (range_min, range_max):
                yaxis_data["data"][yaxis_name] = dict(min=yaxis_min, max=yaxis_max)
        return yaxis_data

    @dash_app.callback(
        [
            Output("group-type-input", "style"),
            Output("group-val-input", "style"),
            Output("bins-input", "style"),
            Output("main-inputs", "className"),
            Output("group-inputs-row", "style"),
        ],
        [
            Input("input-data", "modified_timestamp"),
            Input("map-input-data", "modified_timestamp"),
            Input("candlestick-input-data", "modified_timestamp"),
            Input("treemap-input-data", "modified_timestamp"),
            Input("funnel-input-data", "modified_timestamp"),
            Input("clustergram-input-data", "modified_timestamp"),
            Input("pareto-input-data", "modified_timestamp"),
            Input("histogram-input-data", "modified_timestamp"),
        ],
        [
            State("input-data", "data"),
            State("map-input-data", "data"),
            State("candlestick-input-data", "data"),
            State("treemap-input-data", "data"),
            State("funnel-input-data", "data"),
            State("clustergram-input-data", "data"),
            State("pareto-input-data", "data"),
            State("histogram-input-data", "data"),
        ],
    )
    def main_input_class(
        _ts,
        _ts2,
        _ts3,
        _ts4,
        _ts5,
        _ts6,
        _ts7,
        _ts8,
        inputs,
        map_inputs,
        cs_inputs,
        treemap_inputs,
        funnel_inputs,
        clustergram_inputs,
        pareto_inputs,
        histogram_inputs,
    ):
        return main_inputs_and_group_val_display(
            dict_merge(
                inputs,
                map_inputs,
                cs_inputs,
                treemap_inputs,
                funnel_inputs,
                clustergram_inputs,
                pareto_inputs,
                histogram_inputs,
            )
        )

    @dash_app.callback(
        [
            Output("group-val-dropdown", "options"),
            Output("group-val-dropdown", "value"),
        ],
        [
            Input("chart-tabs", "value"),
            Input("group-dropdown", "value"),
            Input("map-group-dropdown", "value"),
            Input("candlestick-group-dropdown", "value"),
            Input("treemap-group-dropdown", "value"),
            Input("funnel-group-dropdown", "value"),
            Input("clustergram-group-dropdown", "value"),
            Input("pareto-group-dropdown", "value"),
            Input("histogram-group-dropdown", "value"),
        ],
        [State("input-data", "data"), State("group-val-dropdown", "value")],
    )
    def group_values(
        chart_type,
        group_cols,
        map_group_cols,
        cs_group_cols,
        treemap_group_cols,
        funnel_group_cols,
        clustergram_group_cols,
        pareto_group_cols,
        histogram_group_cols,
        inputs,
        prev_group_vals,
    ):
        data_id = inputs["data_id"]
        group_cols = group_cols
        if chart_type == "maps":
            group_cols = map_group_cols
        elif chart_type == "candlestick":
            group_cols = cs_group_cols
        elif chart_type == "treemap":
            group_cols = treemap_group_cols
        elif chart_type == "funnel":
            group_cols = funnel_group_cols
        elif chart_type == "clustergram":
            group_cols = clustergram_group_cols
        elif chart_type == "pareto":
            group_cols = pareto_group_cols
        elif chart_type == "histogram":
            group_cols = histogram_group_cols
        group_cols = make_list(group_cols)
        group_types = get_group_types(inputs, group_cols)
        if "groups" not in group_types:
            return [], None
        group_vals = run_query(
            handle_predefined(data_id),
            inputs.get("query"),
            global_state.get_context_variables(data_id),
        )
        group_vals = build_group_val_options(group_vals, group_cols)
        selections = []
        available_vals = [gv["value"] for gv in group_vals]
        if prev_group_vals is not None:
            selections = [pgv for pgv in prev_group_vals if pgv in available_vals]
        if not len(selections) and len(group_vals) <= MAX_GROUPS:
            selections = available_vals
        return group_vals, selections

    @dash_app.callback(
        Output("popup-content", "children"),
        [Input("url", "pathname"), Input("url", "search")],
    )
    def display_page(pathname, search):
        """
        dash callback which gets called on initial load of each dash page (main & popup)
        """
        dash_app.config.suppress_callback_exceptions = False
        if pathname is None:
            raise DtalePreventUpdate
        params = chart_url_params(search)
        params["data_id"] = params.get("data_id") or get_data_id(pathname)
        df = global_state.get_data(params["data_id"])
        settings = global_state.get_settings(params["data_id"]) or {}
        return html.Div(
            charts_layout(df, settings, **params) + saved_charts.build_layout(),
            className="charts-body pb-0",
        )

    custom_geojson.init_callbacks(dash_app)
    drilldown_modal.init_callbacks(dash_app)
    extended_aggregations.init_callbacks(dash_app)
    lock_zoom.init_callbacks(dash_app)
    saved_charts.init_callbacks(dash_app)



import numpy as np
import pandas as pd

from dash.dependencies import Input, Output, State

import dtale.global_state as global_state
from dtale.charts.utils import AGGS
from dtale.dash_application.charts import (
    build_chart,
    bar_builder,
    build_axes,
    chart_builder_passthru,
)
from dtale.dash_application.exceptions import DtalePreventUpdate
from dtale.dash_application.layout.utils import (
    build_cols,
    build_option,
    build_selections,
)
from dtale.utils import (
    classify_type,
    dict_merge,
    find_dtype,
    is_app_root_defined,
    json_date,
    json_float,
    make_list,
    get_dtypes,
)
from dtale.query import build_query, handle_predefined, run_query
from dtale.charts.utils import (
    MAX_GROUPS,
    ZAXIS_CHARTS,
    build_group_inputs_filter,
    convert_date_val_to_date,
)
from dtale.translations import text


def combine_inputs(dash_app, inputs, chart_inputs={}, yaxis_data={}, map_data={}):
    """
    Combines all managed state (inputs, chart inputs, map inputs & yaxis inputs) within Dash into one dictionary.
    """
    all_inputs = dict_merge(
        inputs, chart_inputs, dict(yaxis=yaxis_data or {}), map_data
    )
    if is_app_root_defined(dash_app.server.config.get("APPLICATION_ROOT")):
        all_inputs["app_root"] = dash_app.server.config["APPLICATION_ROOT"]
    return all_inputs


def build_histogram(data_id, col, query, point_filter):
    data = run_query(
        handle_predefined(data_id), query, global_state.get_context_variables(data_id)
    )
    query, _ = build_group_inputs_filter(data, [point_filter])
    data = run_query(data, query)
    s = data[~pd.isnull(data[col])][col]
    hist_data, hist_labels = np.histogram(s, bins=10)
    hist_labels = list(map(lambda x: json_float(x, precision=3), hist_labels[1:]))
    axes_builder = build_axes(
        dict(
            data=dict(all=dict(Frequency=hist_data, Bins=hist_labels)),
            min=dict(Frequency=0),
            max=dict(Frequency=max(hist_data)),
        ),
        "Bins",
        dict(type="single", data={}),
        data_id=data_id,
    )
    hist_data = dict(data={"all": dict(x=hist_labels, Frequency=hist_data)})
    bars = bar_builder(
        hist_data,
        "Bins",
        ["Frequency"],
        axes_builder,
        chart_builder_passthru,
        modal=True,
    )
    bars.figure["layout"]["xaxis"]["type"] = "category"
    bars.figure["layout"]["title"]["text"] = "{} {} ({} {})".format(
        text("Histogram of"), col, len(s), text("data points")
    )
    return bars


def build_drilldown_title(data_id, all_inputs, click_point, props, val_prop):
    data = global_state.get_data(data_id)

    def _build_val(col, val):
        if classify_type(find_dtype(data[col])) == "D":
            return json_date(convert_date_val_to_date(val))
        return val

    if "text" in click_point:  # Heatmaps
        strs = []
        for dim in click_point["text"].split("<br>"):
            prop, val = dim.split(": ")
            strs.append("{} ({})".format(prop, val))
        return "{}: {}".format(text("Drilldown for"), ", ".join(strs))

    strs = []
    frame_col = all_inputs.get("animate_by")
    if frame_col:
        strs.append("{} ({})".format(frame_col, click_point.get("customdata")))
    for prop in props:
        prop = make_list(prop)
        val_key = prop[0]
        if click_point.get(val_key) is not None:
            col = make_list(all_inputs.get(prop[-1]))[0]
            strs.append(
                "{} ({})".format(col, _build_val(col, click_point.get(val_key)))
            )

    val_prop = make_list(val_prop)
    val_key = val_prop[0]
    val_col = make_list(all_inputs.get(val_prop[-1]))[0]
    agg = AGGS[all_inputs.get("agg") or "raw"]
    strs.append(
        "{} {} ({})".format(agg, val_col, _build_val(val_col, click_point.get(val_key)))
    )
    return "{}: {}".format(text("Drilldown for"), ", ".join(strs))


def init_callbacks(dash_app):
    def toggle_modal(close1, close2, click_data, is_open, inputs, drilldowns_on):
        if not drilldowns_on:
            raise DtalePreventUpdate
        if (inputs.get("agg") or "raw") == "raw":
            raise DtalePreventUpdate
        if close1 or close2 or click_data:
            return not is_open
        return is_open

    def build_x_dropdown(is_open, inputs, chart_inputs, yaxis_data, map_data):
        if not is_open:
            raise DtalePreventUpdate
        data_id = inputs["data_id"]
        df = global_state.get_data(data_id)
        all_inputs = combine_inputs(
            dash_app, inputs, chart_inputs, yaxis_data, map_data
        )
        chart_type, x, y, z, group, map_val, animate_by = (
            all_inputs.get(p)
            for p in ["chart_type", "x", "y", "z", "group", "map_val", "animate_by"]
        )
        if chart_type == "maps":
            if all_inputs.get("map_type") == "choropleth":
                x = all_inputs["loc"]
            else:
                x = "lat_lon"
            x_options = build_selections(map_val, animate_by)
        else:
            x_options = build_selections(z or y, animate_by)
        col_opts = list(build_cols(df.columns, get_dtypes(df)))
        x_options = [build_option(c, l) for c, l in col_opts if c not in x_options]
        return x_options, x

    def update_click_data(
        click_data,
        inputs,
        chart_inputs,
        yaxis_data,
        map_data,
        current_click_data,
        drilldowns_on,
    ):
        if not drilldowns_on:
            raise DtalePreventUpdate
        all_inputs = combine_inputs(
            dash_app, inputs, chart_inputs, yaxis_data, map_data
        )
        if (all_inputs.get("agg") or "raw") == "raw":
            raise DtalePreventUpdate
        if not click_data:
            raise DtalePreventUpdate
        data_id = inputs["data_id"]
        chart_type = all_inputs.get("chart_type")
        click_data = click_data or current_click_data
        click_point = next((p for p in click_data.get("points", [])), None)
        if chart_type == "maps":
            props = [["location", "loc"], "lat", "lon"]
            val = ["z", "map_val"]
        else:
            props = ["x"]
            val = "y"
            if click_point.get("z"):
                props = ["x", "y"]
                val = "z"

        header = build_drilldown_title(data_id, all_inputs, click_point, props, val)
        return click_data, header

    def load_drilldown_content(
        _click_data_ts,
        drilldown_type,
        drilldown_x,
        inputs,
        chart_inputs,
        yaxis_data,
        map_data,
        click_data,
        drilldowns_on,
    ):
        if not drilldowns_on:
            raise DtalePreventUpdate
        data_id = inputs["data_id"]
        all_inputs = combine_inputs(
            dash_app, inputs, chart_inputs, yaxis_data, map_data
        )
        agg = all_inputs.get("agg") or "raw"
        chart_type = all_inputs.get("chart_type")
        frame_col = all_inputs.get("animate_by")
        all_inputs.pop("animate_by", None)
        if agg == "raw":
            raise DtalePreventUpdate
        if drilldown_x is None and chart_type != "maps":
            raise DtalePreventUpdate
        if click_data:
            click_point = next((p for p in click_data.get("points", [])), None)
            if click_point:
                query = build_query(
                    data_id, all_inputs.get("query") or global_state.get_query(data_id)
                )
                x_col = all_inputs.get("x")
                y_col = next((y2 for y2 in make_list(all_inputs.get("y"))), None)
                if chart_type in ZAXIS_CHARTS:
                    x, y, z, frame = (
                        click_point.get(p) for p in ["x", "y", "z", "customdata"]
                    )
                    if chart_type == "heatmap":
                        click_point_vals = {}
                        for dim in click_point["text"].split("<br>"):
                            prop, val = dim.split(": ")
                            click_point_vals[prop] = val
                        x, y, frame = (
                            click_point_vals.get(p) for p in [x_col, y_col, frame_col]
                        )
                    point_filter = {x_col: x, y_col: y}
                    if frame_col:
                        point_filter[frame_col] = frame
                    if drilldown_type == "histogram":
                        z_col = next(
                            (z2 for z2 in make_list(all_inputs.get("z"))), None
                        )
                        hist_chart = build_histogram(
                            data_id, z_col, query, point_filter
                        )
                        return hist_chart, dict(display="none")
                    else:
                        xy_query, _ = build_group_inputs_filter(
                            global_state.get_data(data_id), [point_filter]
                        )
                        if not query:
                            query = xy_query
                        else:
                            query = "({}) and ({})".format(query, xy_query)
                        all_inputs["query"] = query
                        all_inputs["chart_type"] = drilldown_type
                        all_inputs["agg"] = "raw"
                        all_inputs["modal"] = True
                        all_inputs["x"] = drilldown_x
                        all_inputs["y"] = [all_inputs["z"]]
                        chart, _, _, _ = build_chart(**all_inputs)
                        return chart, None

                elif chart_type == "maps":
                    map_type = all_inputs.get("map_type")
                    point_filter = {}
                    if frame_col:
                        point_filter[frame_col] = click_point["customdata"]
                    if map_type == "choropleth":
                        point_filter[all_inputs["loc"]] = click_point["location"]
                    elif map_type == "scattergeo":
                        lat, lon = (click_point.get(p) for p in ["lat", "lon"])
                        point_filter[all_inputs["lat"]] = lat
                        point_filter[all_inputs["lon"]] = lon
                    map_val = all_inputs["map_val"]
                    if drilldown_type == "histogram":
                        hist_chart = build_histogram(
                            data_id, map_val, query, point_filter
                        )
                        return hist_chart, dict(display="none")
                    else:
                        map_query, _ = build_group_inputs_filter(
                            global_state.get_data(data_id), [point_filter]
                        )
                        if not query:
                            query = map_query
                        else:
                            query = "({}) and ({})".format(query, map_query)
                        all_inputs["query"] = query
                        all_inputs["chart_type"] = drilldown_type
                        all_inputs["agg"] = "raw"
                        all_inputs["modal"] = True

                        data = global_state.get_data(data_id)
                        all_inputs["x"] = drilldown_x
                        x_style = None
                        if map_type != "choropleth":
                            all_inputs["x"] = "lat_lon"
                            lat, lon = (all_inputs.get(p) for p in ["lat", "lon"])
                            data.loc[:, "lat_lon"] = (
                                data[lat].astype(str) + "|" + data[lon].astype(str)
                            )
                            x_style = dict(display="none")
                        all_inputs["y"] = [map_val]
                        chart, _, _, _ = build_chart(data=data, **all_inputs)
                        return chart, x_style
                else:
                    x_filter = click_point.get("x")
                    point_filter = {x_col: x_filter}
                    if frame_col:
                        point_filter[frame_col] = click_point.get("customdata")
                    if drilldown_type == "histogram":
                        hist_chart = build_histogram(
                            data_id, y_col, query, point_filter
                        )
                        return hist_chart, dict(display="none")
                    else:
                        x_query, _ = build_group_inputs_filter(
                            global_state.get_data(data_id), [point_filter]
                        )
                        if not query:
                            query = x_query
                        else:
                            query = "({}) and ({})".format(query, x_query)
                        all_inputs["query"] = query
                        all_inputs["chart_type"] = drilldown_type
                        all_inputs["agg"] = "raw"
                        all_inputs["modal"] = True
                        all_inputs["x"] = drilldown_x
                        chart, _, _, _ = build_chart(**all_inputs)
                        return chart, None
        return None, dict(display="none")

    for i in range(1, MAX_GROUPS + 1):
        dash_app.callback(
            Output("drilldown-modal-{}".format(i), "is_open"),
            [
                Input("close-drilldown-modal-{}".format(i), "n_clicks"),
                Input("close-drilldown-modal-header-{}".format(i), "n_clicks"),
                Input("chart-{}".format(i), "clickData"),
            ],
            [
                State("drilldown-modal-{}".format(i), "is_open"),
                State("input-data", "data"),
                State("drilldown-toggle", "on"),
            ],
        )(toggle_modal)
        dash_app.callback(
            [
                Output("chart-click-data-{}".format(i), "data"),
                Output("drilldown-modal-header-{}".format(i), "children"),
            ],
            [Input("chart-{}".format(i), "clickData")],
            [
                State("input-data", "data"),
                State("chart-input-data", "data"),
                State("yaxis-data", "data"),
                State("map-input-data", "data"),
                State("chart-click-data-{}".format(i), "data"),
                State("drilldown-toggle", "on"),
            ],
        )(update_click_data)
        dash_app.callback(
            [
                Output("drilldown-x-dropdown-{}".format(i), "options"),
                Output("drilldown-x-dropdown-{}".format(i), "value"),
            ],
            [Input("drilldown-modal-{}".format(i), "is_open")],
            [
                State("input-data", "data"),
                State("chart-input-data", "data"),
                State("yaxis-data", "data"),
                State("map-input-data", "data"),
            ],
        )(build_x_dropdown)
        dash_app.callback(
            [
                Output("drilldown-content-{}".format(i), "children"),
                Output("drilldown-x-input-{}".format(i), "style"),
            ],
            [
                Input("chart-click-data-{}".format(i), "modified_timestamp"),
                Input("drilldown-chart-type-{}".format(i), "value"),
                Input("drilldown-x-dropdown-{}".format(i), "value"),
            ],
            [
                State("input-data", "data"),
                State("chart-input-data", "data"),
                State("yaxis-data", "data"),
                State("map-input-data", "data"),
                State("chart-click-data-{}".format(i), "data"),
                State("drilldown-toggle", "on"),
            ],
        )(load_drilldown_content)



import dtale.global_state as global_state


def get_data_id(pathname):
    """
    Parses data ID from query path (ex: 'foo/bar/1' => '1')
    """
    data_id = pathname.split("/")[-1]
    data_id = global_state.get_data_id_by_name(data_id) or data_id
    if (global_state.size()) == 0:
        return None
    if global_state.contains(data_id):
        return data_id
    else:
        return sorted(global_state.keys())[0]



import dash_bootstrap_components as dbc

from dtale.dash_application import dcc, html
from dtale.translations import text
from dtale.utils import dict_merge, classify_type, flatten_lists, make_list
from dtale.charts.utils import INDEX_COL
from dtale.constants import CHART_JOINER_CHAR


def show_style(show, display_style="block"):
    return {"display": display_style if show else "none"}


def build_input(
    label, input, className="col-auto", label_class="input-group-addon", **kwargs
):
    """
    Helper function to build a standard label/input component in dash.

    :param label: name of the input you are displaying
    :type label: str
    :param input: dash component for storing state
    :param className: style class to be applied to encapsulating div
    :type className: str
    :param kwargs: Optional keyword arguments to be applied to encapsulating div (style, title, id...)
    :type kwargs: dict
    :return: dash components for label/input
    :rtype: :dash:`dash_html_components.Div <dash-html-components/div>`
    """
    return html.Div(
        [
            html.Div(
                [html.Span(label, className=label_class)] + make_list(input),
                className="input-group mr-3",
            )
        ],
        className=className,
        **kwargs
    )


def build_option(value, label=None):
    """
    Returns value/label inputs in a dictionary for use in
    :dash:`dash_core_components.Dropdown <dash-core-components/Dropdown>`
    """
    return {"label": value if label is None else label, "value": value}


FREQS = ["H", "H2", "WD", "D", "W", "M", "Q", "Y"]
FREQ_LABELS = dict(
    H="Hourly",
    H2="Hour",
    WD="Weekday",
    W="Weekly",
    M="Monthly",
    Q="Quarterly",
    Y="Yearly",
)


def build_cols(cols, dtypes):
    """
    Helper function to add additional column entries for columns of type datetime so that users can make use of
    different frequencies of dates.  For example, hour, weekday, month, quarter, year

    :param cols: columns in dataframe
    :type cols: list of strings
    :param dtypes: datatypes of columns in dataframe
    :type dtypes: dict
    :return: generator or columns + any additional (datetime column + frequency) options
    """
    yield INDEX_COL, INDEX_COL
    for c in cols:
        if classify_type(dtypes[c]) == "D":
            for freq in FREQS:
                if freq in FREQ_LABELS:
                    yield "{}{}{}".format(c, CHART_JOINER_CHAR, freq), "{} ({})".format(
                        c, FREQ_LABELS[freq]
                    )
                else:
                    yield c, c
        else:
            yield c, c


def build_selections(*args):
    """
    simple helper function to build a single level list of values based on variable number of inputs which could be
    equal to None.
    """
    return flatten_lists([[] if a is None else make_list(a) for a in args])


def build_tab(label, value, additional_style=None, **kwargs):
    """
    Builds a :dash:`dash_core_components.Tab <dash-core-components/tab>` with standard styling settings.
    """
    base_style = {"borderBottom": "1px solid #d6d6d6", "padding": "6px"}
    return dcc.Tab(
        label=label,
        value=value,
        style=dict_merge(base_style, {"fontWeight": "bold"}, additional_style or {}),
        disabled_style=dict_merge(
            base_style,
            {
                "fontWeight": "bold",
                "backgroundColor": "LightGray",
                "color": "black",
                "cursor": "not-allowed",
            },
            additional_style or {},
        ),
        selected_style=dict_merge(
            base_style,
            {
                "borderTop": "1px solid #d6d6d6",
                "backgroundColor": "#2a91d1",
                "color": "white",
            },
            additional_style or {},
        ),
        **kwargs
    )


def build_drilldown_modal(idx):
    return dbc.Modal(
        [
            dbc.ModalHeader(
                html.Div(
                    [
                        html.Div(
                            text("Chart Drilldown"),
                            className="col mt-auto mb-auto",
                            id="drilldown-modal-header-{}".format(idx),
                        ),
                        html.Button(
                            html.Span("X"),
                            className="close mr-5",
                            id="close-drilldown-modal-header-{}".format(idx),
                        ),
                    ],
                    className="row",
                )
            ),
            dbc.ModalBody(
                dcc.Loading(
                    [
                        html.Div(
                            [
                                html.Div(
                                    dcc.Tabs(
                                        id="drilldown-chart-type-{}".format(idx),
                                        value="histogram",
                                        children=[
                                            build_tab(t.capitalize(), t)
                                            for t in ["histogram", "bar"]
                                        ],
                                        style=dict(height="36px"),
                                    ),
                                    className="col-md-4",
                                ),
                                build_input(
                                    "X",
                                    dcc.Dropdown(
                                        id="drilldown-x-dropdown-{}".format(idx),
                                        options=[],
                                        placeholder="Select a column",
                                        value=None,
                                        style=dict(width="inherit"),
                                        className="drilldown-x-dropdown",
                                    ),
                                    id="drilldown-x-input-{}".format(idx),
                                    style=dict(display="none"),
                                ),
                            ],
                            className="row pt-3 pb-5",
                        ),
                        html.Div(id="drilldown-content-{}".format(idx)),
                    ],
                    type="circle",
                )
            ),
            dbc.ModalFooter(
                dbc.Button(
                    text("Close"),
                    id="close-drilldown-modal-{}".format(idx),
                    className="ml-auto",
                )
            ),
        ],
        id="drilldown-modal-{}".format(idx),
        size="lg",
        centered=True,
        className="drilldown-modal",
    )


CHART_IDX = 0


def reset_charts():
    global CHART_IDX

    CHART_IDX = 0


def graph_wrapper(modal=False, export=False, **kwargs):
    global CHART_IDX

    curr_style = kwargs.pop("style", None) or {}
    if modal or export:
        id = "{}-chart".format("modal" if modal else "export")
        return dcc.Graph(
            id=id, style=dict_merge({"height": "100%"}, curr_style), **kwargs
        )

    CHART_IDX += 1
    graph = dcc.Graph(
        id="chart-{}".format(CHART_IDX),
        style=dict_merge({"height": "100%"}, curr_style),
        **kwargs
    )
    graph.figure["id"] = "chart-figure-{}".format(CHART_IDX)
    click_data_store = dcc.Store(id="chart-click-data-{}".format(CHART_IDX))
    return [graph, click_data_store, build_drilldown_modal(CHART_IDX)]


def build_hoverable(
    content,
    hoverable_content,
    hover_class="map-types",
    top="50%",
    additional_classes="",
    hover_style=None,
):
    return html.Div(
        [
            content,
            html.Div(
                hoverable_content,
                className="hoverable__content {}".format(hover_class),
                style=dict_merge(dict(top=top), hover_style or dict()),
            ),
        ],
        style=dict(borderBottom="none"),
        className="hoverable {}".format(additional_classes),
    )



import json
import re

import numpy as np
import pandas as pd
from six import string_types

import dtale.global_state as global_state
from dtale.utils import classify_type, find_dtype, format_data, make_list
from dtale.query import build_col_key

# Patterns that could enable code execution via pandas.DataFrame.query()
_DANGEROUS_PATTERNS = re.compile(
    r"(__\w+__"  # dunder attributes (__import__, __class__, etc.)
    r"|(?<!\w)import\s*\("  # import() calls
    r"|(?<!\w)exec\s*\("  # exec() calls
    r"|(?<!\w)eval\s*\("  # eval() calls
    r"|(?<!\w)compile\s*\("  # compile() calls
    r"|(?<!\w)open\s*\("  # open() calls
    r"|(?<!\w)getattr\s*\("  # getattr() calls
    r"|(?<!\w)setattr\s*\("  # setattr() calls
    r"|(?<!\w)delattr\s*\("  # delattr() calls
    r"|(?<!\w)globals\s*\("  # globals() calls
    r"|(?<!\w)locals\s*\("  # locals() calls
    r"|(?<!\w)vars\s*\("  # vars() calls
    r"|(?<!\w)dir\s*\("  # dir() calls
    r"|(?<!\w)type\s*\("  # type() calls
    r"|(?<!\w)os\."  # os module access
    r"|(?<!\w)sys\."  # sys module access
    r"|(?<!\w)subprocess"  # subprocess module access
    r"|(?<!\w)shutil\."  # shutil module access
    r"|@\w+"  # @ references to local variables in query scope
    r")",
    re.IGNORECASE,
)


class ColumnFilterSecurity(object):
    """Validation helpers to guard against code injection via DataFrame.query()."""

    @staticmethod
    def validate_string_value(val):
        """Validate a string value used in filter queries."""
        if not isinstance(val, string_types):
            val = str(val)
        if _DANGEROUS_PATTERNS.search(val):
            raise ValueError(
                "Filter value contains potentially unsafe content: {}".format(repr(val))
            )
        return val

    @staticmethod
    def validate_numeric_value(val):
        """Validate that a value is actually numeric. Returns the original value if valid."""
        if val is None:
            return val
        if isinstance(val, (int, float, np.integer, np.floating)):
            return val
        # Try to parse string as a number (validation only, return original value)
        try:
            if isinstance(val, string_types):
                stripped = val.strip()
                if _DANGEROUS_PATTERNS.search(stripped):
                    raise ValueError(
                        "Numeric filter value contains potentially unsafe content: {}".format(
                            repr(val)
                        )
                    )
                float(stripped)  # validate it parses as a number
                return val  # return original to preserve existing query format
        except (ValueError, TypeError):
            pass
        raise ValueError("Expected numeric value, got: {}".format(repr(val)))

    @staticmethod
    def validate_date_value(val):
        """Validate that a value is a safe date string."""
        if val is None:
            return val
        if not isinstance(val, string_types):
            val = str(val)
        # Date values should only contain digits, dashes, colons, T, Z, dots, spaces, +
        if not re.match(r"^[\d\-/:T Z.+]+$", val):
            raise ValueError(
                "Date filter value contains invalid characters: {}".format(repr(val))
            )
        if _DANGEROUS_PATTERNS.search(val):
            raise ValueError(
                "Date filter value contains potentially unsafe content: {}".format(
                    repr(val)
                )
            )
        return val

    @staticmethod
    def validate_operand(operand, allowed):
        """Validate that the operand is in the allowed set."""
        if operand not in allowed:
            raise ValueError(
                "Invalid operand: {}. Allowed: {}".format(repr(operand), allowed)
            )
        return operand

    @staticmethod
    def validate_outlier_query(query):
        """Validate an outlier filter query string."""
        if query is None:
            return query
        if not isinstance(query, string_types):
            raise ValueError("Outlier query must be a string")
        if _DANGEROUS_PATTERNS.search(query):
            raise ValueError(
                "Outlier filter query contains potentially unsafe content: {}".format(
                    repr(query)
                )
            )
        return query


class ArcticDBColumnFilter(object):
    def __init__(self, saved_filter):
        self.cfg = saved_filter
        column, classification, filter_type = (
            self.cfg["meta"].get(p) for p in ["column", "classification", "type"]
        )
        self.column = column
        self.classification = classification
        if filter_type == "string":
            self.builder = StringFilter(column, self.classification, self.cfg)
        if filter_type in ["int", "float"]:
            self.builder = NumericFilter(column, self.classification, self.cfg)
        if filter_type == "date":
            self.builder = DateFilter(column, self.classification, self.cfg)
        if filter_type == "outliers":
            self.builder = OutlierFilter(column, self.classification, self.cfg)


class ColumnFilter(object):
    def __init__(self, data_id, column, cfg):
        self.data_id = data_id
        self.column = column
        dtype = (global_state.get_dtype_info(data_id, column) or {}).get("dtype")
        if not dtype:
            if global_state.is_arcticdb:
                instance = global_state.store.get(data_id)
                data, _ = format_data(instance.base_df)
                s = data[column]
            else:
                s = global_state.get_data(data_id, columns=[column])[column]
            dtype = find_dtype(s)
        self.classification = classify_type(dtype)
        self.cfg = cfg
        if self.cfg is not None:
            self.cfg = json.loads(self.cfg)
        if self.cfg["type"] == "string":
            self.builder = StringFilter(column, self.classification, self.cfg)
        if self.cfg["type"] in ["int", "float"]:
            self.builder = NumericFilter(column, self.classification, self.cfg)
        if self.cfg["type"] == "date":
            self.builder = DateFilter(column, self.classification, self.cfg)
        if self.cfg["type"] == "outliers":
            self.builder = OutlierFilter(column, self.classification, self.cfg)

    def save_filter(self):
        curr_settings = global_state.get_settings(self.data_id)
        filters_key = "{}Filters".format(
            "outlier" if self.cfg["type"] == "outliers" else "column"
        )
        curr_filters = curr_settings.get(filters_key) or {}
        fltr = self.builder.build_filter()
        if fltr is None:
            curr_filters.pop(self.column, None)
        else:
            curr_filters[self.column] = fltr
        curr_settings[filters_key] = curr_filters
        global_state.set_settings(self.data_id, curr_settings)
        return curr_filters


class OutlierFilter(object):
    def __init__(self, column, classification, cfg):
        self.column = column
        self.classification = classification
        self.cfg = cfg

    def build_filter(self):
        if self.cfg.get("query") is None:
            return None
        ColumnFilterSecurity.validate_outlier_query(self.cfg.get("query"))
        return self.cfg


class MissingOrPopulatedFilter(object):
    def __init__(self, column, classification, cfg):
        self.column = column
        self.classification = classification
        self.cfg = cfg

    def handle_missing_or_populated(self, fltr):
        if self.cfg is None or (
            not self.cfg.get("missing", False) and not self.cfg.get("populated", False)
        ):
            return fltr
        if self.cfg.get("missing", False):
            return {
                "missing": True,
                "meta": {
                    "column": self.column,
                    "classification": self.classification,
                    "type": self.cfg["type"],
                },
                "query": "{col}.isnull()".format(col=build_col_key(self.column)),
            }
        if self.cfg.get("populated", False):
            return {
                "populated": True,
                "meta": {
                    "column": self.column,
                    "classification": self.classification,
                    "type": self.cfg["type"],
                },
                "query": "~{col}.isnull()".format(col=build_col_key(self.column)),
            }

    def update_missing_or_populated_query_builder(self, query_builder, fltr=None):
        if self.cfg is None or (
            not self.cfg.get("missing", False) and not self.cfg.get("populated", False)
        ):
            return fltr
        # TODO: how to handle scenarios where QueryBuilder doesn't support functionality so do it manually
        if self.cfg.get("missing", False):
            return query_builder[self.column] != query_builder[self.column]
        if self.cfg.get("populated", False):
            return query_builder[self.column] == query_builder[self.column]


def handle_ne(query, operand):
    if operand != "=":
        return "~({})".format(query)
    return query


def handle_query_builder_ne(query, operand):
    if operand != "=":
        return ~query
    return query


class StringFilter(MissingOrPopulatedFilter):
    def __init__(self, column, classification, cfg):
        super(StringFilter, self).__init__(column, classification, cfg)

    def build_filter(self):
        if self.cfg is None:
            return super(StringFilter, self).handle_missing_or_populated(None)

        action = self.cfg.get("action", "equals")
        # Validate action
        if action not in (
            "equals",
            "startswith",
            "endswith",
            "contains",
            "regex",
            "length",
        ):
            raise ValueError("Invalid string filter action: {}".format(repr(action)))
        if action == "equals" and not len(self.cfg.get("value", [])):
            return super(StringFilter, self).handle_missing_or_populated(None)
        elif action != "equals" and not self.cfg.get("raw"):
            return super(StringFilter, self).handle_missing_or_populated(None)

        state = self.cfg.get("value", [])
        case_sensitive = self.cfg.get("caseSensitive", False)
        operand = self.cfg.get("operand", "=")
        raw = self.cfg.get("raw")

        # Validate inputs
        ColumnFilterSecurity.validate_operand(operand, {"=", "ne"})
        if raw is not None:
            ColumnFilterSecurity.validate_string_value(raw)
        for v in state:
            ColumnFilterSecurity.validate_string_value(v)

        fltr = dict(
            value=state,
            operand=operand,
            caseSensitive=case_sensitive,
            action=action,
            raw=raw,
            meta={
                "column": self.column,
                "classification": self.classification,
                "type": self.cfg["type"],
            },
        )
        fmt_string = "{!r}" if self.classification == "S" else "{}"
        if action == "equals":
            if len(state) == 1:
                val_str = fmt_string.format(state[0])
                fltr["query"] = "{} {} {}".format(
                    build_col_key(self.column),
                    "==" if operand == "=" else "!=",
                    val_str,
                )
            else:
                val_str = ", ".join(map(fmt_string.format, state))
                fltr["query"] = "{} {} ({})".format(
                    build_col_key(self.column),
                    "in" if operand == "=" else "not in",
                    val_str,
                )
        elif action in ["startswith", "endswith"]:
            case_insensitive_conversion = "" if case_sensitive else ".str.lower()"
            fltr["query"] = "{}{}.str.{}({!r}, na=False)".format(
                build_col_key(self.column),
                case_insensitive_conversion,
                action,
                raw if case_sensitive else raw.lower(),
            )
            fltr["query"] = handle_ne(fltr["query"], operand)
        elif action in ["contains", "regex"]:
            fltr["query"] = "{}.str.contains({!r}, na=False, case={}, regex={})".format(
                build_col_key(self.column),
                raw,
                "True" if case_sensitive else "False",
                "True" if action == "regex" else "False",
            )
            fltr["query"] = handle_ne(fltr["query"], operand)
        elif action == "length":
            # Validate length values are numeric
            if "," in raw:
                start, end = raw.split(",")
                ColumnFilterSecurity.validate_numeric_value(start.strip())
                ColumnFilterSecurity.validate_numeric_value(end.strip())
                fltr["query"] = "{start} <= {col}.str.len() <= {end}".format(
                    col=build_col_key(self.column), start=start, end=end
                )
            else:
                ColumnFilterSecurity.validate_numeric_value(raw.strip())
                fltr["query"] = "{}.str.len() == {}".format(
                    build_col_key(self.column), raw
                )
            fltr["query"] = handle_ne(fltr["query"], operand)
        return super(StringFilter, self).handle_missing_or_populated(fltr)

    def update_query_builder(self, query_builder):
        if self.cfg is None:
            return super(StringFilter, self).update_missing_or_populated_query_builder(
                query_builder
            )

        action = self.cfg.get("action", "equals")
        if action == "equals" and not len(self.cfg.get("value", [])):
            return super(StringFilter, self).update_missing_or_populated_query_builder(
                query_builder
            )
        elif action != "equals" and not self.cfg.get("raw"):
            return super(StringFilter, self).update_missing_or_populated_query_builder(
                query_builder
            )

        state = self.cfg.get("value", [])
        case_sensitive = self.cfg.get("caseSensitive", False)
        operand = self.cfg.get("operand", "=")
        raw = self.cfg.get("raw")
        fltr = dict(
            value=state,
            operand=operand,
            caseSensitive=case_sensitive,
            action=action,
            raw=raw,
        )
        if action == "equals":
            if len(state) == 1:
                fltr["query"] = handle_query_builder_ne(
                    query_builder[self.column] == state[0], operand
                )
            else:
                fltr["query"] = handle_query_builder_ne(
                    query_builder[self.column].isin(state), operand
                )
        elif action in ["startswith", "endswith"]:
            # Not supported by QueryBuilder
            pass
        elif action in ["contains", "regex"]:
            # Not supported by QueryBuilder
            pass
        elif action == "length":
            # Not supported by QueryBuilder
            pass
        return super(StringFilter, self).update_missing_or_populated_query_builder(
            query_builder, fltr.get("query")
        )


class NumericFilter(MissingOrPopulatedFilter):
    def __init__(self, column, classification, cfg):
        super(NumericFilter, self).__init__(column, classification, cfg)

    def build_filter(self):
        if self.cfg is None:
            return super(NumericFilter, self).handle_missing_or_populated(None)
        cfg_val, cfg_operand, cfg_min, cfg_max = (
            self.cfg.get(p) for p in ["value", "operand", "min", "max"]
        )

        # Validate operand (allow None for missing/populated-only filters)
        if cfg_operand is not None:
            ColumnFilterSecurity.validate_operand(
                cfg_operand, {"=", "ne", "<", ">", "<=", ">=", "[]", "()"}
            )

        base_fltr = dict(
            operand=cfg_operand,
            meta={
                "column": self.column,
                "classification": self.classification,
                "type": self.cfg["type"],
            },
        )
        if cfg_operand in ["=", "ne"]:
            state = make_list(cfg_val or [])
            if not len(state):
                return super(NumericFilter, self).handle_missing_or_populated(None)
            # Validate all values are numeric
            state = [ColumnFilterSecurity.validate_numeric_value(v) for v in state]
            fltr = dict(value=cfg_val, **base_fltr)
            if len(state) == 1:
                fltr["query"] = "{} {} {}".format(
                    build_col_key(self.column),
                    "==" if cfg_operand == "=" else "!=",
                    state[0],
                )
            else:
                fltr["query"] = "{} {} ({})".format(
                    build_col_key(self.column),
                    "in" if cfg_operand == "=" else "not in",
                    ", ".join(map(str, state)),
                )
            return super(NumericFilter, self).handle_missing_or_populated(fltr)
        if cfg_operand in ["<", ">", "<=", ">="]:
            if cfg_val is None:
                return super(NumericFilter, self).handle_missing_or_populated(None)
            cfg_val = ColumnFilterSecurity.validate_numeric_value(cfg_val)
            fltr = dict(
                value=cfg_val,
                query="{} {} {}".format(
                    build_col_key(self.column), cfg_operand, cfg_val
                ),
                **base_fltr
            )
            return super(NumericFilter, self).handle_missing_or_populated(fltr)
        if cfg_operand in ["[]", "()"]:
            fltr = dict(**base_fltr)
            queries = []
            if cfg_min is not None:
                cfg_min = ColumnFilterSecurity.validate_numeric_value(cfg_min)
                fltr["min"] = cfg_min
                queries.append(
                    "{} >{} {}".format(
                        build_col_key(self.column),
                        "=" if cfg_operand == "[]" else "",
                        cfg_min,
                    )
                )
            if cfg_max is not None:
                cfg_max = ColumnFilterSecurity.validate_numeric_value(cfg_max)
                fltr["max"] = cfg_max
                queries.append(
                    "{} <{} {}".format(
                        build_col_key(self.column),
                        "=" if cfg_operand == "[]" else "",
                        cfg_max,
                    )
                )
            if len(queries) == 2 and cfg_max == cfg_min:
                queries = ["{} == {}".format(build_col_key(self.column), cfg_max)]
            if not len(queries):
                return super(NumericFilter, self).handle_missing_or_populated(None)
            fltr["query"] = " and ".join(queries)
            return super(NumericFilter, self).handle_missing_or_populated(fltr)
        return super(NumericFilter, self).handle_missing_or_populated(None)

    def update_query_builder(self, query_builder):
        if self.cfg is None:
            return super(NumericFilter, self).update_missing_or_populated_query_builder(
                query_builder
            )
        cfg_val, cfg_operand, cfg_min, cfg_max = (
            self.cfg.get(p) for p in ["value", "operand", "min", "max"]
        )

        if cfg_operand in ["=", "ne"]:
            state = make_list(cfg_val or [])
            if self.cfg.get("meta", {}).get("type") == "float":
                state = [np.float64(val) for val in state]
            if not len(state):
                return super(
                    NumericFilter, self
                ).update_missing_or_populated_query_builder(query_builder)
            fltr = dict(value=cfg_val, operand=cfg_operand)
            if len(state) == 1:
                fltr["query"] = handle_query_builder_ne(
                    query_builder[self.column] == state[0], cfg_operand
                )
            else:
                fltr["query"] = handle_query_builder_ne(
                    query_builder[self.column].isin(state), cfg_operand
                )
            return super(NumericFilter, self).update_missing_or_populated_query_builder(
                query_builder, fltr.get("query")
            )
        if cfg_operand in ["<", ">", "<=", ">="]:
            if cfg_val is None:
                return super(
                    NumericFilter, self
                ).update_missing_or_populated_query_builder(query_builder)
            fltr = dict(value=cfg_val, operand=cfg_operand)
            if cfg_operand == "<":
                fltr["query"] = query_builder[self.column] < cfg_val
            elif cfg_operand == ">":
                fltr["query"] = query_builder[self.column] > cfg_val
            elif cfg_operand == "<=":
                fltr["query"] = query_builder[self.column] <= cfg_val
            elif cfg_operand == ">=":
                fltr["query"] = query_builder[self.column] >= cfg_val
            return super(NumericFilter, self).update_missing_or_populated_query_builder(
                query_builder, fltr.get("query")
            )
        if cfg_operand in ["[]", "()"]:
            # Not supported by QueryBuilder
            return super(NumericFilter, self).update_missing_or_populated_query_builder(
                query_builder
            )
        return super(NumericFilter, self).update_missing_or_populated_query_builder(
            query_builder
        )


class DateFilter(MissingOrPopulatedFilter):
    def __init__(self, column, classification, cfg):
        super(DateFilter, self).__init__(column, classification, cfg)

    def build_filter(self):
        if self.cfg is None:
            return super(DateFilter, self).handle_missing_or_populated(None)

        start, end = (self.cfg.get(p) for p in ["start", "end"])
        # Validate date values
        if start:
            ColumnFilterSecurity.validate_date_value(start)
        if end:
            ColumnFilterSecurity.validate_date_value(end)
        fltr = dict(
            start=start,
            end=end,
            meta={
                "column": self.column,
                "classification": self.classification,
                "type": self.cfg["type"],
            },
        )
        queries = []
        if start:
            queries.append("{} >= '{}'".format(build_col_key(self.column), start))
        if end:
            queries.append("{} <= '{}'".format(build_col_key(self.column), end))
        if len(queries) == 2 and start == end:
            queries = ["{} == '{}'".format(build_col_key(self.column), start)]
        if not len(queries):
            return super(DateFilter, self).handle_missing_or_populated(None)
        fltr["query"] = " and ".join(queries)
        return super(DateFilter, self).handle_missing_or_populated(fltr)

    def update_query_builder(self, query_builder):
        # TODO: need to use datetime.datetime and then for equivalence you need to do (col > input - 1) & (col <= input)
        # pd.Timestamp('2023-01-04').to_pydatetime() -> to get datetime.datetime
        if self.cfg is None:
            return super(DateFilter, self).update_missing_or_populated_query_builder(
                query_builder
            )

        start, end = (self.cfg.get(p) for p in ["start", "end"])
        fltr = dict(start=start, end=end)
        queries = []
        if start:
            start = (
                pd.Timestamp(start) - pd.tseries.frequencies.to_offset("1D")
            ).to_pydatetime()
            queries.append(query_builder[self.column] > start)
        if end:
            queries.append(
                query_builder[self.column] <= pd.Timestamp(end).to_pydatetime()
            )
        if len(queries) == 2 and start == end:
            start_start = (
                pd.Timestamp(start) - pd.tseries.frequencies.to_offset("1D")
            ).to_pydatetime()
            start_end = pd.Timestamp(start).to_pydatetime()
            queries = [
                (query_builder[self.column] > start_start)
                & (query_builder[self.column] <= start_end)
            ]
        if not len(queries):
            return super(DateFilter, self).handle_missing_or_populated(None)
        if len(queries) == 2:
            fltr["query"] = queries[0] & queries[1]
        else:
            fltr["query"] = queries[0]
        return super(DateFilter, self).update_missing_or_populated_query_builder(
            query_builder, fltr.get("query")
        )



import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
import pprint
import scipy.stats as sts
import time

import dtale.global_state as global_state
import dtale.pandas_util as pandas_util

from dtale.code_export import build_code_export, build_final_chart_code
from dtale.column_builders import clean, clean_code
from dtale.describe import load_describe
from dtale.query import build_query, load_filterable_data
from dtale.utils import (
    apply,
    classify_type,
    find_dtype,
    find_dtype_formatter,
    find_selected_column,
    get_bool_arg,
    get_int_arg,
    get_str_arg,
    grid_columns,
    grid_formatter,
    json_float,
    json_timestamp,
    make_list,
)

LINE_CFG = "line={'shape': 'spline', 'smoothing': 0.3}, mode='lines'"


def handle_top(df, top):
    if top is not None:
        top = int(top)
        if top > 0:
            return df[:top], top, ["chart = chart[:{}]".format(top)]
        else:
            return df[top:], top, ["chart = chart[{}:]".format(top)]
    elif len(df) > 100:
        top = 100
        return df[:top], top, ["chart = chart[:100]"]
    return df, len(df), []


def pctsum_updates(df, group, ret_col):
    grp_sums = df.groupby(group)[[ret_col]].sum()
    code = (
        "ordinal_data = df.groupby('{col}')[['{ret_col}']].sum()\n"
        "ordinal_data = ordinal_data / ordinal_data.sum()"
    ).format(col=group, ret_col=ret_col)
    return grp_sums / grp_sums.sum(), [code]


def handle_cleaners(s, cleaners):
    cleaner_code = []
    if cleaners:
        for cleaner in cleaners.split(","):
            s = clean(s, cleaner, {})
            cleaner_code += clean_code(cleaner, {})
    return s, cleaner_code


def build_kde(s, hist_labels, selected_col):
    try:
        kde = sts.gaussian_kde(s)
        kde_data = kde.pdf(hist_labels)
        kde_data = [json_float(k, precision=12) for k in kde_data]
        code = [
            "import scipy.stats as sts",
            "kde = sts.gaussian_kde(s['{}'])".format(selected_col),
            "kde_data = kde.pdf(np.linspace(labels.min(), labels.max()))",
        ]
        return kde_data, code
    except np.linalg.LinAlgError:
        return None, []


class ColumnAnalysis(object):
    def __init__(self, data_id, req):
        self.data_id = data_id
        self.analysis_type = get_str_arg(req, "type")
        self.query = build_query(data_id, global_state.get_query(data_id))
        self.data = load_filterable_data(data_id, req, query=self.query)
        self.selected_col = find_selected_column(
            self.data, get_str_arg(req, "col", "values")
        )
        if self.analysis_type != "frequency":
            self.data = self.data[~pd.isnull(self.data[self.selected_col])]
        self.dtype = find_dtype(self.data[self.selected_col])
        self.classifier = classify_type(self.dtype)
        self.code = build_code_export(
            data_id,
            imports="{}\n".format(
                "\n".join(
                    [
                        "import numpy as np",
                        "import pandas as pd",
                        "import plotly.graph_objs as go",
                    ]
                )
            ),
        )

        if self.analysis_type is None:
            self.analysis_type = (
                "histogram" if self.classifier in ["F", "I", "D"] else "value_counts"
            )

        if self.analysis_type == "geolocation":
            self.analysis = GeolocationAnalysis(req)
        elif self.analysis_type == "histogram":
            self.analysis = HistogramAnalysis(req)
        elif self.analysis_type == "categories":
            self.analysis = CategoryAnalysis(req)
        elif self.analysis_type == "value_counts":
            self.analysis = ValueCountAnalysis(req)
        elif self.analysis_type == "word_value_counts":
            self.analysis = WordValueCountAnalysis(req)
        elif self.analysis_type == "qq":
            self.analysis = QQAnalysis()
        elif self.analysis_type == "frequency":
            self.analysis = FrequencyAnalysis(req)

    def build(self):
        base_code = build_code_export(
            self.data_id,
            imports="{}\n\n".format(
                "\n".join(
                    [
                        "import numpy as np",
                        "import pandas as pd",
                        "import plotly.graph_objs as go",
                    ]
                )
            ),
        )
        return_data, code = self.analysis.build(self)
        return dict(
            code=build_final_chart_code(base_code + code),
            query=self.query,
            cols=global_state.get_dtypes(self.data_id),
            dtype=self.dtype,
            chart_type=self.analysis_type,
            timestamp=round(time.time() * 1000),
            **return_data
        )


class HistogramAnalysis(object):
    def __init__(self, req):
        self.bins = get_int_arg(req, "bins", 20)
        self.target = get_str_arg(req, "target")
        self.density = get_bool_arg(req, "density")

    def build_histogram_data(self, series):
        hist_kwargs = {"density": True} if self.density else {"bins": self.bins}
        hist_data, hist_labels = np.histogram(series, **hist_kwargs)
        hist_data = [json_float(h) for h in hist_data]
        decimals = 1
        # drop the first bin because of just a minimum
        labels = [
            "".join(["{0:.", "{}".format(decimals), "f}"]).format(lbl)
            for lbl in hist_labels[1:]
        ]
        while len(set(labels)) < len(labels) or decimals > 10:
            decimals += 1
            labels = [
                "".join(["{0:.", "{}".format(decimals), "f}"]).format(lbl)
                for lbl in hist_labels[1:]
            ]

        return dict(labels=labels, data=hist_data), hist_labels

    def build(self, parent):
        if parent.classifier == "D":
            pandas_util.assign_col_data(
                parent.data,
                parent.selected_col,
                apply(parent.data[parent.selected_col], json_timestamp),
            )
        kde_code = []
        if self.target is None:
            return_data, hist_labels = self.build_histogram_data(
                parent.data[parent.selected_col]
            )
            kde, kde_code = build_kde(
                parent.data[parent.selected_col], hist_labels, parent.selected_col
            )
            if kde is not None:
                return_data["kde"] = kde
        else:
            bin_vals = pd.cut(parent.data[parent.selected_col], bins=self.bins)
            labels = ["{}".format(c) for c in bin_vals.dtype.categories]
            pandas_util.assign_col_data(parent.data, "bin", bin_vals.astype("str"))
            return_data = {"targets": [], "labels": labels}
            target_dtype = find_dtype(parent.data[self.target])
            target_formatter = find_dtype_formatter(target_dtype)
            for target, target_data in parent.data[[self.target, "bin"]].groupby(
                self.target
            ):
                target_counts = target_data["bin"].value_counts()
                target_counts = [
                    int(tc) for tc in target_counts.reindex(labels, fill_value=0).values
                ]
                return_data["targets"].append(
                    dict(
                        target=target_formatter(target, as_string=True),
                        data=target_counts,
                    )
                )

        desc, desc_code = load_describe(parent.data[parent.selected_col])
        dtype_info = global_state.get_dtype_info(parent.data_id, parent.selected_col)
        for p in ["skew", "kurt"]:
            if p in dtype_info:
                desc[p] = dtype_info[p]

        return_data["desc"] = desc
        return return_data, self._build_code(parent, kde_code, desc_code)

    def _build_code(self, parent, kde_code, desc_code):
        pp = pprint.PrettyPrinter(indent=4)

        code = [
            "s = df[~pd.isnull(df['{col}'])][['{col}'{target}]]".format(
                col=parent.selected_col,
                target=",'{}'".format(self.target) if self.target else "",
            )
        ]
        if parent.classifier == "D":
            code.append(
                (
                    "\nimport time\n\n"
                    "s.loc[:, '{col}'] = s['{col}'].apply(\n"
                    "\tlambda x: int((time.mktime(x.timetuple()) + (old_div(x.microsecond, 1000000.0))) * 1000\n"
                    ")"
                ).format(col=parent.selected_col)
            )
        if self.target is None:
            hist_kwargs = (
                "density=True" if self.density else "bins={}".format(self.bins)
            )
            code.append(
                "chart, labels = np.histogram(s['{col}'], {hist_kwargs})".format(
                    col=parent.selected_col, hist_kwargs=hist_kwargs
                )
            )
            code += kde_code + desc_code
            layout = pp.pformat(
                go.Layout(
                    **{
                        "barmode": "group",
                        "legend": {"orientation": "h"},
                        "title": {
                            "text": "{} Histogram (bins: {}) w/ KDE".format(
                                parent.selected_col, self.bins
                            )
                        },
                        "xaxis2": {"anchor": "y", "overlaying": "x", "side": "top"},
                        "yaxis": {"title": {"text": "Frequency"}, "side": "left"},
                        "yaxis2": {
                            "title": {"text": "KDE"},
                            "side": "right",
                            "overlaying": "y",
                        },
                    }
                )
            )
            code += [
                "charts = [",
                "\tgo.Bar(x=labels[1:], y=chart, name='Histogram'),",
                "\tgo.Scatter(",
                "\t\tx=list(range(len(kde_data))), y=kde_data, name='KDE',"
                "\t\tyaxis='y2', xaxis='x2',"
                "\t\t{}".format(LINE_CFG),
                "\t)",
                "]",
            ]
        else:
            layout = pp.pformat(
                go.Layout(
                    **{
                        "barmode": "stack",
                        "legend": {"orientation": "h"},
                        "title": {
                            "text": "{} Histogram (bins: {}) w/ target ({})".format(
                                parent.selected_col, self.bins, self.target
                            )
                        },
                        "yaxis": {"title": {"text": "Frequency"}, "side": "left"},
                    }
                )
            )
            code.append(
                (
                    "bin_vals = pd.cut(s['{col}'], bins={bins})\n"
                    "labels = [str(c) for c in bin_vals.dtype.categories]\n"
                    "s.loc[:, 'bin'] = bin_vals.astype('str')\n"
                    "for target, target_data in s[['{target}', 'bin']].groupby('{target}'):\n"
                    "\ttarget_counts = target_data['bin'].value_counts()\n"
                    "\ttarget_counts = [int(tc) for tc in target_counts.reindex(labels, fill_value=0).values]\n"
                    "\tcharts.append(go.Bar(x=labels, y=target_counts, name=target))"
                ).format(col=parent.selected_col, bins=self.bins, target=self.target)
            )
        code.append(
            "figure = go.Figure(data=charts, layout=go.{layout})".format(layout=layout)
        )
        return code


class CategoryAnalysis(object):
    def __init__(self, req):
        self.category_col = get_str_arg(req, "categoryCol")
        self.category_agg = get_str_arg(req, "categoryAgg", "mean")
        self.aggs = [
            "count",
            "sum" if self.category_agg == "pctsum" else self.category_agg,
        ]
        self.top = get_int_arg(req, "top")

    def build(self, parent):
        hist = pandas_util.groupby(parent.data, self.category_col, dropna=False)[
            [parent.selected_col]
        ].agg(self.aggs)
        hist.columns = hist.columns.droplevel(0)
        hist.columns = ["count", "data"]
        if self.category_agg == "pctsum":
            hist["data"] = hist["data"] / hist["data"].sum()
        hist.index.name = "labels"
        hist = hist.reset_index()
        hist, top, top_code = handle_top(hist, self.top)
        f = grid_formatter(grid_columns(hist), nan_display="NaN")
        return_data = f.format_lists(hist)
        return_data["top"] = top
        return return_data, self._build_code(parent, top_code)

    def _build_code(self, parent, top_code):
        pp = pprint.PrettyPrinter(indent=4)
        layout = pp.pformat(
            go.Layout(
                **{
                    "barmode": "group",
                    "legend": {"orientation": "h"},
                    "title": {
                        "text": "{}({}) Categorized by {}".format(
                            parent.selected_col, self.category_agg, self.category_col
                        )
                    },
                    "xaxis": {"title": {"text": self.category_col}},
                    "yaxis": {
                        "title": {
                            "text": "{} ({})".format(
                                parent.selected_col, self.category_agg
                            )
                        },
                        "side": "left",
                    },
                    "yaxis2": {
                        "title": {"text": "Frequency"},
                        "side": "right",
                        "overlaying": "y",
                    },
                }
            )
        )

        code = [
            "chart = df{groupby}[['{col}']].agg(['{aggs}'])".format(
                groupby=pandas_util.groupby_code(make_list(self.category_col), False),
                col=parent.selected_col,
                aggs="', '".join(self.aggs),
            ),
            "chart.columns = chart.columns.droplevel(0)",
            'chart.columns = ["count", "data"]',
        ]
        if self.category_agg == "pctsum":
            code.append("chart['data'] = chart['data'] / chart['data'].sum()")
        code += ["chart.index.name = 'labels'", "chart = chart.reset_index()"]
        code += top_code
        code += [
            "charts = [",
            "\tgo.Bar(x=chart['labels'].values, y=chart['data'].values),",
            "\tgo.Scatter(",
            "\t\tx=chart['labels'].values, y=chart['count'].values, yaxis='y2',",
            "\t\tname='Frequency', {}".format(LINE_CFG),
            "\t)",
            "]",
            "figure = go.Figure(data=charts, layout=go.{layout})".format(layout=layout),
        ]
        return code


def build_hist(s, code, df_var="chart"):
    df = s.value_counts()
    code.append("{} = s.value_counts().to_frame(name='data')".format(df_var))
    df = df.to_frame(name="data")
    df["percent"] = (df["data"] / df["data"].sum()) * 100
    code.append(
        "{df_var}['percent'] = ({df_var}['data'] / {df_var}['data'].sum()) * 100".format(
            df_var=df_var
        )
    )
    return df


class ValueCountAnalysis(object):
    def __init__(self, req):
        self.top = get_int_arg(req, "top")
        self.ordinal_col = get_str_arg(req, "ordinalCol")
        self.ordinal_agg = get_str_arg(req, "ordinalAgg", "sum")
        self.cleaners = get_str_arg(req, "cleaner")

    def build_hist(self, s, code):
        return build_hist(s, code)

    def setup_ordinal_data(self, parent):
        if self.ordinal_agg == "pctsum":
            return pctsum_updates(parent.data, parent.selected_col, self.ordinal_col)

        ordinal_data = getattr(
            parent.data.groupby(parent.selected_col)[[self.ordinal_col]],
            self.ordinal_agg,
        )()
        ordinal_code = [
            "ordinal_data = df.groupby('{col}')[['{ordinal}']].{agg}()".format(
                col=parent.selected_col, ordinal=self.ordinal_col, agg=self.ordinal_agg
            )
        ]
        return ordinal_data, ordinal_code

    def setup_chart_layout(self, parent):
        pp = pprint.PrettyPrinter(indent=4)
        layout_cfg = {
            "barmode": "group",
            "legend": {"orientation": "h"},
            "title": {"text": "{} Value Counts".format(parent.selected_col)},
            "xaxis": {"title": {"text": parent.selected_col}},
            "yaxis": {"title": {"text": "Frequency"}},
        }
        if self.ordinal_col:
            layout_cfg["yaxis2"] = {
                "title": {"text": "{} ({})".format(self.ordinal_col, self.ordinal_agg)},
                "side": "right",
                "overlaying": "y",
            }
        return pp.pformat(go.Layout(**layout_cfg))

    def build(self, parent):
        code = [
            "s = df[~pd.isnull(df['{col}'])]['{col}']".format(col=parent.selected_col)
        ]
        s, cleaner_code = handle_cleaners(
            parent.data[parent.selected_col], self.cleaners
        )
        code += cleaner_code
        hist = self.build_hist(s, code)

        if self.ordinal_col is not None:
            ordinal_data, ordinal_code = self.setup_ordinal_data(parent)
            code += ordinal_code
            hist["ordinal"] = ordinal_data
            hist.index.name = "labels"
            hist = hist.reset_index().sort_values("ordinal")
            code += [
                "chart['ordinal'] = ordinal_data",
                "chart.index.name = 'labels'",
                "chart = chart.reset_index().sort_values('ordinal')",
            ]
        else:
            hist.index.name = "labels"
            hist = hist.reset_index().sort_values(
                ["data", "labels"], ascending=[False, True]
            )
            code += [
                "chart.index.name = 'labels'",
                "chart = chart.reset_index().sort_values(['data', 'labels'], ascending=[False, True])",
            ]
        hist, top, top_code = handle_top(hist, self.top)
        code += top_code
        col_types = grid_columns(hist)
        f = grid_formatter(col_types, nan_display=None)
        return_data = f.format_lists(hist)
        return_data["top"] = top

        layout = self.setup_chart_layout(parent)
        code.append(
            "charts = [go.Bar(x=chart['labels'].values, y=chart['data'].values, name='Frequency')]"
        )
        if self.ordinal_col:
            code.append(
                (
                    "charts.append(go.Scatter(\n"
                    "\tx=chart['labels'].values, y=chart['ordinal'].values, yaxis='y2',\n"
                    "\tname='{} ({})', {}\n"
                    "))"
                ).format(self.ordinal_col, self.ordinal_agg, LINE_CFG)
            )
        code.append(
            "figure = go.Figure(data=charts, layout=go.{layout})".format(layout=layout)
        )
        return return_data, code


class WordValueCountAnalysis(ValueCountAnalysis):
    def build_hist(self, s, code):
        code.append("chart = s.str.split(expand=True).stack().value_counts()")
        code.append("chart = chart.to_frame(name='data').sort_index()")
        code.append("chart['percent'] = (chart['data'] / chart['data'].sum()) * 100")
        df = (
            s.str.split(expand=True)
            .stack()
            .value_counts()
            .to_frame(name="data")
            .sort_index()
        )
        df["percent"] = (df["data"] / df["data"].sum()) * 100
        return df

    def setup_ordinal_data(self, parent):
        expanded_words = parent.data[parent.selected_col].str.split(expand=True).stack()
        expanded_words.name = "label"
        expanded_words = expanded_words.reset_index()[["level_0", "label"]]
        expanded_words.columns = ["index", "label"]
        expanded_words = pd.merge(
            parent.data[[self.ordinal_col]],
            expanded_words.set_index("index"),
            how="inner",
            left_index=True,
            right_index=True,
        )
        expanded_word_code = [
            (
                "ordinal_data = df['{col}'].str.split(expand=True).stack()\n"
                "ordinal_data.name = 'label'\n"
                "ordinal_data = ordinal_data.reset_index()[['level_0', 'label']]\n"
                "ordinal_data.columns = ['index', 'label']\n"
                "ordinal_data = pd.merge(\n"
                "\tdf[['{ordinal}']],\n"
                "\tordinal_data.set_index('index'),\n"
                "\thow='inner',\n"
                "\tleft_index=True,\n"
                "\tright_index=True,\n"
                ")"
            ).format(col=parent.selected_col, ordinal=self.ordinal_col)
        ]
        if self.ordinal_agg == "pctsum":
            ordinal_data, ordinal_code = pctsum_updates(
                expanded_words, "label", self.ordinal_col
            )
            return ordinal_data, expanded_word_code + ordinal_code

        ordinal_code = expanded_word_code + [
            "ordinal_data = ordinal_data.groupby('label')[['{ordinal}']].{agg}()".format(
                ordinal=self.ordinal_col, agg=self.ordinal_agg
            )
        ]
        ordinal_data = getattr(
            expanded_words.groupby("label")[[self.ordinal_col]], self.ordinal_agg
        )()
        return ordinal_data, ordinal_code

    def setup_chart_layout(self, parent):
        pp = pprint.PrettyPrinter(indent=4)
        layout_cfg = {
            "barmode": "group",
            "legend": {"orientation": "h"},
            "title": {"text": "{} Word Value Counts".format(parent.selected_col)},
            "xaxis": {"title": {"text": parent.selected_col}},
            "yaxis": {"title": {"text": "Frequency"}},
        }
        if self.ordinal_col:
            layout_cfg["yaxis2"] = {
                "title": {"text": "{} ({})".format(self.ordinal_col, self.ordinal_agg)},
                "side": "right",
                "overlaying": "y",
            }
        return pp.pformat(go.Layout(**layout_cfg))


class GeolocationAnalysis(object):
    def __init__(self, req):
        self.lat_col = get_str_arg(req, "latCol")
        self.lon_col = get_str_arg(req, "lonCol")

    def build(self, parent):
        geo = parent.data[[self.lat_col, self.lon_col]].dropna()
        geo.columns = ["lat", "lon"]
        col_types = grid_columns(geo)
        f = grid_formatter(col_types, nan_display=None)
        return_data = f.format_lists(geo)
        return return_data, self._build_code()

    def _build_code(self):
        pp = pprint.PrettyPrinter(indent=4)
        layout = pp.pformat(
            go.Layout(
                **{
                    "autosize": True,
                    "geo": {"fitbounds": "locations", "scope": "world"},
                    "legend": {"orientation": "h"},
                    "margin": {"b": 0, "l": 0, "r": 0},
                    "title": {
                        "text": "Map of Latitude({})/ Longitude({})".format(
                            self.lat_col, self.lon_col
                        )
                    },
                }
            )
        )
        return [
            "chart = df[['{}', '{}']].dropna()".format(self.lat_col, self.lon_col),
            "chart.columns = ['lat', 'lon']",
            (
                "chart = go.Scattergeo(\n"
                "\tlon=chart['lon'].values,\n"
                "\tlat=chart['lat'].values,\n"
                "\tmode='markers',\n"
                "\tmarker={'color': 'darkblue'}\n"
                ")"
            ),
            "figure = go.Figure(data=chart, layout=go.{layout})".format(layout=layout),
        ]


class QQAnalysis(object):
    def build(self, parent):
        s = parent.data[parent.selected_col]
        if parent.classifier == "D":
            s = apply(s, json_timestamp)

        qq_x, qq_y = sts.probplot(s, dist="norm", fit=False)
        qq = pd.DataFrame(dict(x=qq_x, y=qq_y))
        f = grid_formatter(grid_columns(qq), nan_display=None)
        return_data = f.format_lists(qq)

        trend_line = px.scatter(x=qq_x, y=qq_y, trendline="ols").data[1]
        trend_line = pd.DataFrame(dict(x=trend_line["x"], y=trend_line["y"]))
        f = grid_formatter(grid_columns(trend_line), nan_display=None)
        trend_line = f.format_lists(trend_line)
        return_data["x2"] = trend_line["x"]
        return_data["y2"] = trend_line["y"]
        return return_data, self._build_code(parent)

    def _build_code(self, parent):
        pp = pprint.PrettyPrinter(indent=4)
        layout = pp.pformat(
            go.Layout(
                **{
                    "legend": {"orientation": "h"},
                    "title": {"text": "{} QQ Plot".format(parent.selected_col)},
                }
            )
        )
        code = [
            "s = df[~pd.isnull(df['{col}'])]['{col}']".format(col=parent.selected_col)
        ]
        if parent.classifier == "D":
            code.append(
                (
                    "\nimport time\n\n"
                    "s = s['{col}'].apply(\n"
                    "\tlambda x: int((time.mktime(x.timetuple()) + (old_div(x.microsecond, 1000000.0))) * 1000\n"
                    ")"
                ).format(col=parent.selected_col)
            )
        code += [
            "\nimport scipy.stats as sts\nimport plotly.express as px\n",
            'qq_x, qq_y = sts.probplot(s, dist="norm", fit=False)',
            "chart = px.scatter(x=qq_x, y=qq_y, trendline='ols', trendline_color_override='red')",
            "figure = go.Figure(data=chart, layout=go.{layout})".format(layout=layout),
        ]
        return code


"""
http://anschonfel.hn.res.num:9205/dtale/column-analysis/1?col=str_val&top=10&type=frequency&filtered=false
"""


class FrequencyAnalysis(object):
    def __init__(self, req):
        self.top = get_int_arg(req, "top")
        self.split_cols = get_str_arg(req, "splits", "")
        if self.split_cols:
            self.split_cols = self.split_cols.split(",")
        else:
            self.split_cols = []
        self.cleaners = get_str_arg(req, "cleaner")

    def build(self, parent):
        code = []
        if parent.classifier == "S":
            code.append(
                "s = df.fillna('Missing')['{col}']".format(col=parent.selected_col)
            )
            s, cleaner_code = handle_cleaners(
                parent.data[parent.selected_col].fillna("Missing"), self.cleaners
            )
            code += cleaner_code
        else:
            code.append(
                "s = df['{col}'].fillna('Missing').astype(str)".format(
                    col=parent.selected_col
                )
            )
            formatter = find_dtype_formatter(parent.dtype)
            s = parent.data[parent.selected_col].apply(
                lambda x: formatter(x, nan_display="Missing")
            )

        df_var = "base_vals" if len(self.split_cols) else "result"

        base_vals = build_hist(s, code, df_var)
        base_vals.index.name = parent.selected_col
        base_vals = base_vals.rename(
            columns={"data": "Frequency", "percent": "Percent"}
        )
        base_vals = base_vals[base_vals["Frequency"] > 0]
        base_vals = base_vals.reset_index().sort_values(
            ["Frequency", parent.selected_col], ascending=[False, True]
        )
        base_vals = base_vals.head(self.top)

        code += [
            "{}.index.name = '{}'".format(df_var, parent.selected_col),
            "{df_var} = {df_var}.{rename}".format(
                df_var=df_var,
                rename="rename(columns={'data': 'Frequency', 'percent': 'Percent'})",
            ),
            "{df_var} = {df_var}[{df_var}['Frequency'] > 0]".format(df_var=df_var),
            "{df_var} = {df_var}.reset_index().sort_values(['Frequency', '{col}'], ascending=[False, True])".format(
                df_var=df_var, col=parent.selected_col
            ),
            "{df_var} = {df_var}.head(self.top)".format(df_var=df_var),
        ]

        if len(self.split_cols):
            top_vals = base_vals[parent.selected_col]
            val_filter = parent.data[parent.selected_col].isin(top_vals)
            val_filter_code = "val_filter = df['{col}'].isin(top_vals)".format(
                col=parent.selected_col
            )
            if (top_vals == "Missing").any():
                val_filter = val_filter | parent.data[parent.selected_col].isnull()
                val_filter_code = "({val_filter} | df['{col}'].isnull())".format(
                    val_filter=val_filter, col=parent.selected_col
                )
            hist = parent.data[val_filter].groupby([s] + self.split_cols).size()
            hist.name = "Frequency"
            hist = hist.reset_index()
            hist = hist[hist["Frequency"] > 0]
            outer_freq = hist.groupby(parent.selected_col)["Frequency"].transform("sum")
            hist["Percent"] = (hist["Frequency"] / outer_freq) * 100

            code += [
                "top_vals = {df_var}['{col}']".format(
                    df_var=df_var, col=parent.selected_col
                ),
                val_filter_code,
                "result = df[val_filter].groupby([s, '{}']).size()".format(
                    "', '".join(self.split_cols)
                ),
                "result.name = 'Frequency'",
                "result = result.reset_index()",
                "result = result[result['Frequency'] > 0]",
                "outer_freq = result.groupby('{}')['Frequency'].transform('sum')".format(
                    parent.selected_col
                ),
                "result['Percent'] = (result['Frequency'] / outer_freq) * 100",
            ]
        else:
            hist = base_vals

        col_types = grid_columns(hist)
        f = grid_formatter(col_types, nan_display=None)
        return_data = f.format_lists(hist)
        return_data = dict(data=return_data)
        return_data["top"] = self.top
        return return_data, code



import os
import platform
import subprocess

_system = platform.system()
IS_WINDOWS = _system == "Windows"
IS_DARWIN = _system == "Darwin"
IS_LINUX_OR_BSD = (_system == "Linux") or ("BSD" in _system)


def is_executable_in_path(name):
    """Check if executable is in OS path."""
    from distutils.spawn import find_executable

    return find_executable(name) is not None


def _open_browser_with_webbrowser(url):
    import webbrowser

    webbrowser.open(url)


def _open_browser_with_command(command, url):
    cmd_line = [command, url]
    with open(os.devnull, "w") as devnull:
        subprocess.Popen(cmd_line, stdout=devnull, stderr=subprocess.STDOUT)


def open_browser(url):
    """Open a web browser pointing to a given URL.

    We use this function instead of Python's `webbrowser` module because this
    way we can capture stdout/stderr to avoid polluting the terminal with the
    browser's messages. For example, Chrome always prints things like "Created
    new window in existing browser session", and those get on the user's way.

    url : str
        The URL. Must include the protocol.

    """

    # Treat Windows separately because:
    # 1. /dev/null doesn't exist.
    # 2. subprocess.Popen(['start', url]) doesn't actually pop up the
    #    browser even though 'start url' works from the command prompt.
    # Fun!
    # Also, use webbrowser if we are on Linux and xdg-open is not installed.
    #
    # We don't use the webbrowser module on Linux and Mac because some browsers
    # (ahem... Chrome) always print "Opening in existing browser session" to
    # the terminal, which is spammy and annoying. So instead we start the
    # browser ourselves and send all its output to /dev/null.

    if IS_WINDOWS:
        _open_browser_with_webbrowser(url)
        return
    if IS_LINUX_OR_BSD:
        if is_executable_in_path("xdg-open"):
            _open_browser_with_command("xdg-open", url)
            return
        _open_browser_with_webbrowser(url)
        return
    if IS_DARWIN:
        _open_browser_with_command("open", url)
        return

    import platform

    raise Error('Cannot open browser in platform "%s"' % platform.system())


class Error(Exception):
    pass



try:
    from dash import dcc  # noqa: F401
    from dash import html  # noqa: F401

    if not hasattr(html, "Div"):
        import dash_core_components as dcc  # noqa: F811
        import dash_html_components as html  # noqa: F811
except ImportError:
    import dash_core_components as dcc  # noqa: F401
    import dash_html_components as html  # noqa: F401



import pandas as pd

from logging import getLogger

from dtale.utils import parse_version

logger = getLogger(__name__)


def check_pandas_version(version_number):
    return parse_version(pd.__version__) >= parse_version(version_number)


def has_dropna():
    return check_pandas_version("1.1.0")


def groupby(df, index, dropna=True):
    if has_dropna():
        return df.groupby(index, dropna=dropna)
    return df.groupby(index)


def groupby_code(index, dropna=True):
    if has_dropna():
        return ".groupby(['{index}'], dropna={dropna})".format(
            index="','".join(index), dropna=dropna
        )
    return ".groupby(['{index}'])".format(index="','".join(index))


def is_pandas2():
    return check_pandas_version("2.0.0")


def is_pandas3():
    return check_pandas_version("3.0.0")


def run_function(obj, calc_name):
    try:
        if hasattr(obj, calc_name):
            return getattr(obj, calc_name)()
    except BaseException as ex:
        logger.debug("Could not execute {} function: {}".format(calc_name, ex))
    return None


def assign_col_data(df, col, updates):
    if is_pandas2():
        df[col] = updates
    else:
        df.loc[:, col] = updates



import warnings

from flask import Blueprint

with warnings.catch_warnings():
    warnings.filterwarnings(
        "ignore",
        (
            "\nThe dash_html_components package is deprecated. Please replace"
            "\n`import dash_html_components as html` with `from dash import html`"
        ),
    )
    warnings.filterwarnings("ignore", module="matplotlib\\..*")  # noqa: W605

    dtale = Blueprint("dtale", __name__, url_prefix="/dtale")

    ALLOW_CELL_EDITS = True
    HIDE_SHUTDOWN = False
    GITHUB_FORK = False
    HIDE_HEADER_EDITOR = False
    LOCK_HEADER_MENU = False
    HIDE_HEADER_MENU = False
    HIDE_MAIN_MENU = False
    HIDE_COLUMN_MENUS = False
    HIDE_ROW_EXPANDERS = False
    ENABLE_CUSTOM_FILTERS = False
    ENABLE_WEB_UPLOADS = False

    # flake8: NOQA
    from dtale.app import show, get_instance, instances, offline_chart  # isort:skip
    from dtale.cli.loaders import LOADERS  # isort:skip
    from dtale.cli.clickutils import retrieve_version
    from dtale.global_state import update_id  # isort:skip

    for loader_name, loader in LOADERS.items():
        if hasattr(loader, "show_loader"):
            globals()["show_{}".format(loader_name)] = loader.show_loader

    __version__ = retrieve_version("dtale")



import json
import os as _os

import dtale.global_state as global_state
from dtale.utils import read_file

_basepath = _os.path.dirname(__file__)
_languages = {}
for filename in _os.listdir(_basepath):
    lang, ext = _os.path.splitext(filename)
    if ext == ".json":
        filepath = _os.path.abspath(_os.path.join(_basepath, filename))
        _languages[lang] = json.loads(read_file(filepath))


def text(key):
    curr_lang = global_state.get_app_settings()["language"]
    return _languages.get(curr_lang, {}).get(key) or key



#!/usr/bin/env python
from __future__ import print_function

from setuptools import find_packages, setup


def read_file(path):
    # if this fails on windows then add the following environment variable (PYTHONUTF8=1)
    with open(path) as contents:
        return contents.read()


# Convert Markdown to RST for PyPI
# http://stackoverflow.com/a/26737672
try:
    import pypandoc

    pypandoc_func = (
        pypandoc.convert_file if hasattr(pypandoc, "convert_file") else pypandoc.convert
    )
    long_description = pypandoc_func("DESCRIPTION.md", "rst")
except (IOError, ImportError, OSError):
    long_description = read_file("DESCRIPTION.md")


setup(
    name="dtale",
    version="3.21.0",
    author="MAN Alpha Technology",
    author_email="ManAlphaTech@man.com",
    description="Web Client for Visualizing Pandas Objects",
    license="LGPL",
    long_description=long_description,
    keywords=["numeric", "pandas", "visualization", "flask"],
    url="https://github.com/man-group/dtale",
    install_requires=read_file("requirements.txt").splitlines(),
    extras_require={
        "arctic": ["arctic <= 1.79.4"],
        "arcticdb": ["arcticdb"],
        "dash-bio": [
            "ParmEd==3.4.3; python_version == '3.6'",
            "dash-bio; python_version > '3.0'",
            "dash-bio==0.7.1; python_version == '2.7'",
        ],
        "ngrok": ["flask-ngrok; python_version > '3.0'"],
        "r": ["rpy2; python_version > '3.0'"],
        "redis": read_file("requirements-redis.txt").splitlines(),
        "streamlit": ["streamlit"],
        "swifter": ["swifter"],
        "tests": read_file("requirements-test.txt").splitlines(),
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.14",
    ],
    packages=find_packages(exclude=["tests*", "script*"]),
    package_data={
        "dtale": [
            "dash_application/components/*",
            "static/dist/*",
            "static/dash/*",
            "static/css/*",
            "static/fonts/*",
            "static/images/*",
            "static/images/**/*",
            "static/maps/*",
            "templates/**/*",
            "templates/**/**/*",
            "translations/*",
        ]
    },
    entry_points={
        "console_scripts": [
            "dtale = dtale.cli.script:main",
            "dtale-streamlit = dtale.cli.streamlit_script:main",
        ]
    },
    zip_safe=False,
)



import numpy as np

from dtale.utils import grid_columns, grid_formatter, json_int, json_float


def load_describe(column_series, additional_aggs=None):
    """
    Helper function for grabbing the output from :meth:`pandas:pandas.Series.describe` in a JSON serializable format

    :param column_series: data to describe
    :type column_series: :class:`pandas:pandas.Series`
    :return: JSON serializable dictionary of the output from calling :meth:`pandas:pandas.Series.describe`
    """
    desc = column_series.describe().to_frame().T
    code = [
        "# main statistics",
        "stats = df['{col}'].describe().to_frame().T".format(col=column_series.name),
    ]
    if additional_aggs:
        for agg in additional_aggs:
            if agg == "mode":
                mode = column_series.mode().values
                desc["mode"] = np.nan if len(mode) > 1 else mode[0]
                code.append(
                    (
                        "# mode\n"
                        "mode = df['{col}'].mode().values\n"
                        "stats['mode'] = np.nan if len(mode) > 1 else mode[0]"
                    ).format(col=column_series.name)
                )
                continue
            desc[agg] = getattr(column_series, agg)()
            code.append(
                "# {agg}\nstats['{agg}'] = df['{col}'].{agg}()".format(
                    col=column_series.name, agg=agg
                )
            )
    desc_f_overrides = {
        "I": lambda f, i, c: f.add_int(i, c, as_string=True),
        "F": lambda f, i, c: f.add_float(i, c, precision=4, as_string=True),
    }
    desc_f = grid_formatter(
        grid_columns(desc), nan_display="nan", overrides=desc_f_overrides
    )
    desc = desc_f.format_dict(next(desc.itertuples(), None))
    if "count" in desc:
        # pandas always returns 'count' as a float and it adds useless decimal points
        desc["count"] = desc["count"].split(".")[0]
    desc["total_count"] = json_int(len(column_series), as_string=True)
    missing_ct = column_series.isnull().sum()
    desc["missing_pct"] = json_float((missing_ct / len(column_series) * 100).round(2))
    desc["missing_ct"] = json_int(missing_ct, as_string=True)
    return desc, code



from logging import getLogger

import pandas as pd

from dtale.app import show
from dtale.cli.clickutils import get_loader_options

logger = getLogger(__name__)

"""
  IMPORTANT!!! This global variable is required for building any customized CLI loader.
  When find loaders on startup it will search for any modules containing the global variable LOADER_KEY.
"""
NOW = pd.Timestamp("now").strftime("%Y%m%d")
LOADER_KEY = "arcticdb"
LOADER_PROPS = [
    dict(name="uri", help="arctic URI"),
    dict(name="library", help="library within --arcticdb-uri"),
    dict(name="symbol", help="symbol within --arcticdb-library"),
    dict(
        name="start",
        help="start-date of range to load if reading from ChunkStore (EX: {})".format(
            NOW
        ),
    ),
    dict(
        name="end",
        help="end-date of range to load if reading from ChunkStore (EX: {})".format(
            NOW
        ),
    ),
    dict(
        name="use_store",
        help="Use this arcticdb host as the storage mechanism for D-Tale",
        is_flag=True,
    ),
]


# IMPORTANT!!! This function is required if you would like to be able to use this loader from the back-end.
def show_loader(**kwargs):
    return show(data_loader=lambda: loader_func(**kwargs), **kwargs)


def loader_func(**kwargs):
    import dtale.global_state as global_state

    try:
        from arcticdb import Arctic
        from arcticdb.version_store._store import VersionedItem
    except ImportError:
        raise ImportError(
            "In order to use the arcticdb loader you must install arcticdb!"
        )

    read_kwargs = {}
    uri, library, symbol, start, end = (
        kwargs.get(p) for p in ["uri", "library", "symbol", "start", "end"]
    )
    uri = kwargs.get("arcticdb_host") or uri
    if not uri:
        raise ValueError("--arcticdb-uri is a required parameter!")
    if start and end:
        read_kwargs["date_range"] = [pd.Timestamp(start), pd.Timestamp(end)]
    conn = Arctic(uri)

    if kwargs.get("use_store"):
        global_state.use_arcticdb_store(
            **dict(uri=uri, library=library, symbol=symbol, **read_kwargs)
        )

    if global_state.is_arcticdb:
        from dtale.views import startup

        if library and global_state.store.lib.name != library:
            global_state.store.update_library(library)

        if symbol is None:  # select symbol from the UI
            return None

        if library is None and symbol is not None:
            raise ValueError(
                "When trying to load the symbol, {}, a library must be specified!".format(
                    symbol
                )
            )

        startup(data="{}|{}".format(library, symbol))
        return symbol

    if not library:
        raise ValueError(
            "Please specify a value for --arcticdb-library or use --arcticdb-use_store"
        )
    lib = conn.get_library(library)
    data = lib.read(symbol, **read_kwargs)
    if isinstance(data, VersionedItem):
        data = data.data

    return data


# IMPORTANT!!! This function is required for building any customized CLI loader.
def find_loader(kwargs):
    """
    Arctic implementation of data loader which will return a function if any of the
    `click` options based on LOADER_KEY & LOADER_PROPS have been used, otherwise return None

    :param kwargs: Optional keyword arguments to be passed from `click`
    :return: data loader function for arctic implementation
    """
    arctic_opts = get_loader_options(LOADER_KEY, LOADER_PROPS, kwargs)
    if len([f for f in arctic_opts.values() if f]):

        def _arctic_loader():
            return loader_func(**arctic_opts)

        return _arctic_loader
    return None



import pandas as pd

import dtale.global_state as global_state
from dtale.query import run_query
from dtale.utils import dict_merge, grid_columns, grid_formatter, triple_quote
from dtale.charts.utils import build_group_inputs_filter


class NoDuplicatesException(ValueError):
    """Container class for any instance where a user tries to remove duplicates when they don't exist."""


class NoDuplicatesToShowException(ValueError):
    """Container class for any instance where a user tries to show duplicates that don't exist."""


class RemoveAllDataException(ValueError):
    """Container class for any instance where a user tries remove duplicates and it returns an empty dataframe."""


class DuplicateCheck(object):
    def __init__(self, data_id, check_type, cfg):
        self.data_id = data_id
        if check_type == "columns":
            self.checker = DuplicateColumns(data_id, cfg)
        elif check_type == "column_names":
            self.checker = DuplicateColumnNames(data_id, cfg)
        elif check_type == "rows":
            self.checker = RemoveDuplicateRows(data_id, cfg)
        elif check_type == "show":
            self.checker = ShowDuplicates(cfg)
        else:
            raise NotImplementedError(
                "'{}' duplicate check not implemented yet!".format(check_type)
            )

    def test(self):
        data = global_state.get_data(self.data_id)
        return self.checker.check(data)

    def execute(self):
        from dtale.views import startup

        data = global_state.get_data(self.data_id)
        try:
            df, code = self.checker.remove(data)
            instance = startup(data=df, **self.checker.startup_kwargs)
            curr_settings = global_state.get_settings(instance._data_id)
            global_state.set_settings(
                instance._data_id, dict_merge(curr_settings, dict(startup_code=code))
            )
            return instance._data_id
        except NoDuplicatesException:
            return self.data_id


def process_keep(input_list, keep):
    if keep == "first":
        return input_list[1:]
    elif keep == "last":
        return input_list[:-1]
    return input_list  # none


class DuplicateColumns(object):
    def __init__(self, data_id, cfg=None):
        self.cfg = cfg
        self.startup_kwargs = dict(ignore_duplicate=True, data_id=data_id)

    def check(self, df):
        duplicate_columns = {}
        keep = self.cfg.get("keep") or "none"
        col_indexes = list(range(df.shape[1]))
        if keep == "last":
            col_indexes = col_indexes[::-1]
        for x_idx, x in enumerate(col_indexes):
            col = df.iloc[:, x]
            col_duplicates = duplicate_columns.get(df.columns.values[x], [])
            for y in col_indexes[x_idx + 1 :]:
                other_col = df.iloc[:, y]
                if col.equals(other_col):
                    col_duplicates.append(df.columns.values[y])
            duplicate_columns[df.columns.values[x]] = col_duplicates
        return {k: v for k, v in duplicate_columns.items() if len(v) > 0}

    def remove(self, df):
        duplicate_cols = self.check(df)
        keep = self.cfg.get("keep") or "none"
        cols_to_remove = []
        for col, dupes in duplicate_cols.items():
            if keep == "none":
                cols_to_remove += [col] + dupes
            else:
                cols_to_remove += dupes
        if not cols_to_remove:
            raise NoDuplicatesException()
        if len(cols_to_remove) == len(df.columns):
            raise RemoveAllDataException("This will remove all data!")
        df = df[[c for c in df.columns if c not in cols_to_remove]]
        code = (
            "duplicate_cols_to_remove = [\n"
            "\t'{cols}'\n"
            "]\n"
            "df = df[[c for c in df.columns if c not in duplicates_cols_to_remove]]"
        ).format(cols="','".join(cols_to_remove))
        return df, code


class DuplicateColumnNames(object):
    def __init__(self, data_id, cfg=None):
        self.cfg = cfg
        self.startup_kwargs = dict(ignore_duplicate=True, data_id=data_id)

    def check(self, df):
        distinct_names = {}
        for col in df.columns:
            general_name = col.strip().lower()
            names = distinct_names.get(general_name, [])
            names.append(col)
            distinct_names[general_name] = names
        return {k: v for k, v in distinct_names.items() if len(v) > 1}

    def remove(self, df):
        duplicate_names = self.check(df)
        keep = self.cfg.get("keep") or "none"
        names_to_remove = []
        for _, v in duplicate_names.items():
            names_to_remove += process_keep(v, keep)
        if not names_to_remove:
            raise NoDuplicatesException()
        if len(names_to_remove) == len(df.columns):
            raise RemoveAllDataException("This will remove all data!")
        df = df[[c for c in df.columns if c not in names_to_remove]]
        code = (
            "duplicate_cols_to_remove = [\n"
            "\t'{cols}'\n"
            "]\n"
            "df = df[[c for c in df.columns if c not in duplicates_cols_to_remove]]"
        ).format(cols="','".join(names_to_remove))
        return df, code


class RemoveDuplicateRows(object):
    def __init__(self, data_id, cfg):
        self.cfg = cfg
        self.startup_kwargs = dict(ignore_duplicate=True, data_id=data_id)

    def check(self, df):
        subset, keep = (self.cfg.get(p) for p in ["subset", "keep"])
        subset = None if not len(subset or []) else subset
        dupe_args = {"keep": False if keep == "none" else keep}
        duplicates = df.duplicated(subset, **dupe_args)
        removed = int(duplicates.sum())
        total = len(df)
        return dict(removed=removed, total=total, remaining=total - removed)

    def remove(self, df):
        subset, keep = (self.cfg.get(p) for p in ["subset", "keep"])
        dupe_args = {"keep": False if keep == "none" else keep}
        duplicates = df.duplicated(subset, **dupe_args)
        dupe_ct = int(duplicates.sum())
        if not dupe_ct:
            raise NoDuplicatesException()
        if dupe_ct == len(df):
            raise RemoveAllDataException("This will remove all data!")
        df = df[~duplicates].reset_index(drop=True)
        code = self._build_code()
        return df, code

    def _build_code(self):
        subset, keep = (self.cfg.get(p) for p in ["subset", "keep"])
        keep = "False" if keep == "none" else "'{}'".format(keep)
        return (
            "duplicates = df.duplicated(['{subset}'], keep={keep})\n"
            "df = df[~duplicates]"
        ).format(subset="','".join(subset), keep=keep)


class ShowDuplicates(object):
    def __init__(self, cfg):
        self.cfg = cfg
        self.startup_kwargs = dict(ignore_duplicate=True, data_id=None)

    def check(self, df):
        group = self.cfg.get("group")
        duplicates = df[group].reset_index().groupby(group).count()
        duplicates = duplicates.iloc[:, 0]
        duplicates = duplicates[duplicates > 1]
        duplicate_counts = duplicates.values
        duplicates = duplicates.reset_index()[group]
        duplicates = grid_formatter(
            grid_columns(duplicates), as_string=True
        ).format_lists(duplicates)
        check_data = {
            ", ".join([duplicates[col][i] for col in group]): dict(
                count=int(ct), filter=[duplicates[col][i] for col in group]
            )
            for i, ct in enumerate(duplicate_counts)
        }
        return check_data

    def remove(self, df):
        group = self.cfg.get("group")
        duplicates = [g for _, g in df.groupby(group) if len(g) > 1]
        if not duplicates:
            raise NoDuplicatesToShowException("No duplicates to show!")
        duplicates = pd.concat(duplicates)
        group_filter = None
        if self.cfg.get("filter"):
            group_filter, _ = build_group_inputs_filter(
                df, [{col: val for col, val in zip(group, self.cfg["filter"])}]
            )
            duplicates = run_query(duplicates, group_filter)
        code = self._build_code(group_filter)
        return duplicates, code

    def _build_code(self, group_filter=None):
        group = self.cfg.get("group")
        group_filter_str = ""
        if group_filter:
            group_filter_str = "\ndf = df.query({filter})".format(
                filter=triple_quote(group_filter_str)
            )
        return (
            "df = pd.concat(g for _, g in df.groupby(['{group}']) if len(g) > 1)"
            "{filter}"
        ).format(group="','".join(group), filter=group_filter_str)
